# -*- coding: utf-8 -*-

#  Copyright 2023 Quarkslab
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import sqlite3

from .types import Element, ElementComponent, ElementComponentType, Edge, \
    EdgeType, Node, NodeType, Symbol, SymbolType, File, FileContent, \
    LocalSymbol, SourceLocation, SourceLocationType, Occurrence, Error, \
    ComponentAccess, ComponentAccessType


# ------------------------------------------------------------------------ #
# Data Access Object (DAO) for Sourcetrail database                        #
# ------------------------------------------------------------------------ #

class SqliteHelper(object):
    """
        Helper class for sqlite operation
    """

    @staticmethod
    def connect(path: str) -> sqlite3.Connection:
        """
            Wrapper for sqlite3 connect method so the api doesn't rely 
            directly on sqlite and his more general 
            :param path: The path to the database, if the path doesn't point
            to an existing file, a new database file will be created
            :return: A connection handle that can be used for future
            operation on the database
        """
        return sqlite3.connect(path)

    @staticmethod
    def exec(database: sqlite3.Connection, request: str,
             parameters: tuple = ()) -> int:
        """
            Execute the sqlite request without returning the result 
            :param database: A database handle
            :param request: The SQL request to execute
            :param parameters: A tuple containing values for the bind
            parameters of the SQL request (if any)
            :return: The id of the last modified row (useful in case insertion)
        """

        if not database:
            raise Exception('Invalid database handle')

        cur = database.cursor()
        cur.execute(request, parameters)
        cur.close()
        return cur.lastrowid

    @staticmethod
    def fetch(database: sqlite3.Connection, request: str,
              parameters: tuple = ()) -> list:
        """
            Return the result of the sqlite request as list 
            :param database: A database handle
            :param request: The SQL request to execute
            :param parameters: A tuple containing values for the bind
            parameters of the SQL request (if any)
            :return: A list containing the results of the SQL request 
        """

        if not database:
            raise Exception('Invalid database handle')

        cur = database.cursor()
        cur.execute(request, parameters)
        result = cur.fetchall()
        cur.close()

        return result


class ElementDAO(object):
    """
        This class is a static class that can manipulate Element objects,
        inserting and removing them from a sqlite database.
    """

    @staticmethod
    def create_table(database: sqlite3.Connection) -> None:
        """
            Create the element table of the Sourcetrail database
            if it doesn't exist.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''    
            CREATE TABLE IF NOT EXISTS element(
                id INTEGER,                         
                PRIMARY KEY(id));'''
        )

    @staticmethod
    def delete_table(database: sqlite3.Connection) -> None:
        """
            Delete the element table of the Sourcetrail database
            only if it exists.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''
            DROP TABLE IF EXISTS main.element;'''
        )

    @staticmethod
    def new(database: sqlite3.Connection, obj: Element) -> int:
        """
            Insert a new Element inside the element table.
            :param database: A database handle
            :param obj: The object to insert
            :return: The id of the inserted element
        """
        # The 'obj' parameter is not used because we need to create a identifier
        # that does not already exists in the database. So instead, we insert 
        # NULL and the database does the rest.  
        # The 'obj' parameter is present to add some consistency with the 
        # rest of the API and also for futur proof consideration.
        return SqliteHelper.exec(database, '''
            INSERT INTO element(id) VALUES (NULL);'''
        )

    @staticmethod
    def delete(database: sqlite3.Connection, obj: Element) -> None:
        """
            Delete an Element from the element table.
            :param database: A database handle
            :param obj: The object to delete
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM element WHERE id = ?;''', (obj.id,)
        )

    @staticmethod
    def clear(database: sqlite3.Connection) -> None:
        """
            Delete all Elements from the element table.
            :param database: A database handle
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM element;'''
        )

    @staticmethod
    def get(database: sqlite3.Connection, elem_id: int) -> Element:
        """
            Return an element from the database with the matching id
            :param database: A database handle
            :param elem_id: The id of the element to retrieve  
            :return: A Element object that reflect the content inside 
            the database 
        """
        out = SqliteHelper.fetch(database, '''
            SELECT * FROM element WHERE id = ?;''', (elem_id,)
        )

        if len(out) == 1:
            return Element(*out[0])

    @staticmethod
    def update(database: sqlite3.Connection, obj: Element) -> None:
        """
            Update an Element inside the element table.
            :param database: A database handle
            :param obj: The Element object to update 
            :return: None 
        """
        # Since the Element object does only contain a primary key
        # it can't be updated
        pass

    @staticmethod
    def list(database: sqlite3.Connection) -> list[Element]:
        """
            Return the list of all the elements from the element table. 
            :param database: A database handle
            :return: The list of Elements 
        """
        rows = SqliteHelper.fetch(database, '''
            SELECT * FROM element;'''
        )

        result = list()
        for row in rows:
            result.append(Element(*row))

        return result


class ElementComponentDAO(object):
    """
        This class is a static class that can manipulate ElementComponent objects,
        inserting and removing them from a sqlite database.
    """

    @staticmethod
    def create_table(database: sqlite3.Connection) -> None:
        """
            Create the element_component table of the Sourcetrail database
            if it doesn't exist.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''    
            CREATE TABLE IF NOT EXISTS element_component(
                id INTEGER, 
                element_id INTEGER, 
                type INTEGER, 
                data TEXT, 
                PRIMARY KEY(id), 
                FOREIGN KEY(element_id) REFERENCES element(id) ON DELETE CASCADE
            );'''
        )

    @staticmethod
    def delete_table(database: sqlite3.Connection) -> None:
        """
            Delete the element_component table of the Sourcetrail database
            only if it exists.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''
            DROP TABLE IF EXISTS main.element_component;'''
        )

    @staticmethod
    def new(database: sqlite3.Connection, obj: ElementComponent) -> int:
        """
            Insert a new ElementComponent inside the element_component table.
            :param database: A database handle
            :param obj: The object to insert
            :return: The id of the inserted element
        """
        return SqliteHelper.exec(database, '''
            INSERT INTO element_component(
                id, element_id, type, data
            ) VALUES(NULL, ?, ?, ?);''', (obj.elem_id, obj.type.value, obj.data)
        )

    @staticmethod
    def delete(database: sqlite3.Connection, obj: ElementComponent) -> None:
        """
            Delete an ElementComponent from the element_component table.
            :param database: A database handle
            :param obj: The object to delete
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM element_component WHERE id = ?;''', (obj.id,)
        )

    @staticmethod
    def clear(database: sqlite3.Connection) -> None:
        """
            Delete all ElementComponents from the element_component table.
            :param database: A database handle
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM element_component;'''
        )

    @staticmethod
    def get(database: sqlite3.Connection, elem_id: int) -> ElementComponent:
        """
            Return a ElementComponent from the database with the matching id 
            :param database: A database handle
            :param elem_id: The id of the element_component to retrieve  
            :return: A ElementComponent object that reflect the content 
            inside the database 
        """
        out = SqliteHelper.fetch(database, '''
            SELECT * FROM element_component WHERE id = ?;''', (elem_id,)
        )

        if len(out) == 1:
            id_, element_id, type_, data = out[0]
            return ElementComponent(id_, element_id,
                ElementComponentType(type_), data
        )

    @staticmethod
    def update(database: sqlite3.Connection, obj: ElementComponent) -> None:
        """
            Update an ElementComponent inside the element_component table.
            :param database: A database handle
            :param obj: The Element object to update 
            :return: None 
        """
        SqliteHelper.exec(database, '''
            UPDATE element_component SET
                element_id = ?, 
                type = ?, 
                data = ?
            WHERE
                id = ?;''', (obj.elem_id, obj.type.value, obj.data, obj.id)
        )

    @staticmethod
    def list(database: sqlite3.Connection) -> list[ElementComponent]:
        """
            Return the list of all the elements from the element_component table. 
            :param database: A database handle
            :return: The list of ElementComponents 
        """
        rows = SqliteHelper.fetch(database, '''
            SELECT * FROM element_component;'''
        )

        result = list()
        for row in rows:
            id_, element_id, type_, data = row
            result.append(ElementComponent(
                id_, element_id, ElementComponentType(type_), data
            ))

        return result


class EdgeDAO(object):
    """
        This class is a static class that can manipulate Edge objects,
        inserting and removing them from a sqlite database.
    """

    @staticmethod
    def create_table(database: sqlite3.Connection) -> None:
        """
            Create the edge table of the Sourcetrail database
            if it doesn't exist.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''    
            CREATE TABLE IF NOT EXISTS edge(
                id INTEGER NOT NULL, 
                type INTEGER NOT NULL, 
                source_node_id INTEGER NOT NULL, 
                target_node_id INTEGER NOT NULL, 
                PRIMARY KEY(id), 
                FOREIGN KEY(id) REFERENCES element(id) ON DELETE CASCADE, 
                FOREIGN KEY(source_node_id) REFERENCES node(id) ON DELETE CASCADE, 
                FOREIGN KEY(target_node_id) REFERENCES node(id) ON DELETE CASCADE
            );'''
        )

    @staticmethod
    def delete_table(database: sqlite3.Connection) -> None:
        """
            Delete the edge table of the Sourcetrail database
            only if it exists.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''
            DROP TABLE IF EXISTS main.edge;'''
        )

    @staticmethod
    def new(database: sqlite3.Connection, obj: Edge) -> int:
        """
            Insert a new Edge inside the edge table.
            :param database: A database handle
            :param obj: The object to insert
            :return: The id of the inserted element
        """
        return SqliteHelper.exec(database, '''
            INSERT INTO edge(
                id, type, source_node_id, target_node_id
            ) VALUES(?, ?, ?, ?);''', (obj.id, obj.type.value, obj.src, obj.dst)
        )

    @staticmethod
    def delete(database: sqlite3.Connection, obj: Edge) -> None:
        """
            Delete an Edge from the edge table.
            :param database: A database handle
            :param obj: The object to delete
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM edge WHERE id = ?;''', (obj.id,)
        )

    @staticmethod
    def clear(database: sqlite3.Connection) -> None:
        """
            Delete all Edges from the edge table.
            :param database: A database handle
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM edge;'''
        )

    @staticmethod
    def get(database: sqlite3.Connection, elem_id: int) -> Edge:
        """
            Return an Edge from the database with the matching id 
            :param database: A database handle
            :param elem_id: The id of the element to retrieve  
            :return: A Edge object that reflect the content inside 
            the database 
        """
        out = SqliteHelper.fetch(database, '''
            SELECT * FROM edge WHERE id = ?;''', (elem_id,)
        )

        if len(out) == 1:
            id_, type_, src, dst = out[0]
            return Edge(id_, EdgeType(type_), src, dst)

    @staticmethod
    def update(database: sqlite3.Connection, obj: Edge) -> None:
        """
            Update an Edge inside the element table.
            :param database: A database handle
            :param obj: The Edge object to update 
            :return: None 
        """
        SqliteHelper.exec(database, '''
            UPDATE edge SET
                type = ?, 
                source_node_id = ?,
                target_node_id = ?
            WHERE
                id = ?;''', (obj.type.value, obj.src, obj.dst, obj.id)
        )

    @staticmethod
    def list(database: sqlite3.Connection) -> list[Edge]:
        """
            Return the list of all the elements from the edge table. 
            :param database: A database handle
            :return: The list of Edges 
        """
        rows = SqliteHelper.fetch(database, '''
            SELECT * FROM edge;'''
        )

        result = list()
        for row in rows:
            id_, type_, src, dst = row
            result.append(Edge(
                id_, EdgeType(type_), src, dst
            ))

        return result


class NodeDAO(object):
    """
        This class is a static class that can manipulate Node objects,
        inserting and removing them from a sqlite database.
    """

    @staticmethod
    def create_table(database: sqlite3.Connection) -> None:
        """
            Create the node table of the Sourcetrail database
            if it doesn't exist.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''    
            CREATE TABLE IF NOT EXISTS node(
                id INTEGER NOT NULL, 
                type INTEGER NOT NULL, 
                serialized_name TEXT, 
                PRIMARY KEY(id), 
                FOREIGN KEY(id) REFERENCES element(id) ON DELETE CASCADE
            );'''
        )

    @staticmethod
    def delete_table(database: sqlite3.Connection) -> None:
        """
            Delete the node table of the Sourcetrail database
            only if it exists.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''
            DROP TABLE IF EXISTS main.node;'''
        )

    @staticmethod
    def new(database: sqlite3.Connection, obj: Node) -> int:
        """
            Insert a new Node inside the node table.
            :param database: A database handle
            :param obj: The object to insert
            :return: The id of the inserted element
        """
        return SqliteHelper.exec(database, '''
            INSERT INTO node(
                id, type, serialized_name 
            ) VALUES(?, ?, ?);''', (obj.id, obj.type.value, obj.name)
        )

    @staticmethod
    def delete(database: sqlite3.Connection, obj: Node) -> None:
        """
            Delete a Node from the node table.
            :param database: A database handle
            :param obj: The object to delete
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM node WHERE id = ?;''', (obj.id,)
        )

    @staticmethod
    def clear(database: sqlite3.Connection) -> None:
        """
            Delete all Nodes from the node table.
            :param database: A database handle
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM node;'''
        )

    @staticmethod
    def get(database: sqlite3.Connection, elem_id: int) -> Node:
        """
            Return a Node from the database with the matching id 
            :param database: A database handle
            :param elem_id: The id of the element to retrieve  
            :return: A Node object that reflect the content inside 
            the database 
        """
        out = SqliteHelper.fetch(database, '''
            SELECT * FROM node WHERE id = ?;''', (elem_id,)
        )

        if len(out) == 1:
            id_, type_, serialized_name = out[0]
            return Node(id_, NodeType(type_), serialized_name)

    @staticmethod
    def get_by_name(database: sqlite3.Connection, name: str) -> Node:
        """
            Return a Node from the database with the matching serialized_name 
            :param database: A database handle
            :param name: The serialized_name of the element to retrieve  
            :return: A Node object that reflect the content inside 
            the database 
        """
        out = SqliteHelper.fetch(database, '''
            SELECT * FROM node WHERE serialized_name = ? LIMIT 1;''', (name,)
        )

        if len(out) == 1:
            id_, type_, serialized_name = out[0]
            return Node(id_, NodeType(type_), serialized_name)

    @staticmethod
    def update(database: sqlite3.Connection, obj: Node) -> None:
        """
            Update a Node inside the node table.
            :param database: A database handle
            :param obj: The Node object to update 
            :return: None 
        """
        SqliteHelper.exec(database, '''
            UPDATE node SET
                type = ?, 
                serialized_name = ?
            WHERE
                id = ?;''', (obj.type.value, obj.name, obj.id)
        )

    @staticmethod
    def list(database: sqlite3.Connection) -> list[Node]:
        """
            Return the list of all the elements from the node table. 
            :param database: A database handle
            :return: The list of Nodes 
        """
        rows = SqliteHelper.fetch(database, '''
            SELECT * FROM node;'''
        )

        result = list()
        for row in rows:
            id_, type_, serialized_name = row
            result.append(Node(id_, NodeType(type_), serialized_name))

        return result


class SymbolDAO(object):
    """
        This class is a static class that can manipulate Symbol objects,
        inserting and removing them from a sqlite database.
    """

    @staticmethod
    def create_table(database: sqlite3.Connection) -> None:
        """
            Create the symbol table of the Sourcetrail database
            if it doesn't exist.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''    
            CREATE TABLE symbol(
                id INTEGER NOT NULL, 
                definition_kind INTEGER NOT NULL, 
                PRIMARY KEY(id), 
                FOREIGN KEY(id) REFERENCES node(id) ON DELETE CASCADE
            );'''
        )

    @staticmethod
    def delete_table(database: sqlite3.Connection) -> None:
        """
            Delete the symbol table of the Sourcetrail database
            only if it exists.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''
            DROP TABLE IF EXISTS main.symbol;'''
        )

    @staticmethod
    def new(database: sqlite3.Connection, obj: Symbol) -> int:
        """
            Insert a new Symbol inside the symbol table.
            :param database: A database handle
            :param obj: The object to insert
            :return: The id of the inserted symbol
        """
        return SqliteHelper.exec(database, '''
            INSERT INTO symbol(
                id, definition_kind 
            ) VALUES(?, ?);''', (obj.id, obj.definition_kind.value)
        )

    @staticmethod
    def delete(database: sqlite3.Connection, obj: Symbol) -> None:
        """
            Delete a Symbol from the symbol table.
            :param database: A database handle
            :param obj: The object to delete
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM symbol WHERE id = ?;''', (obj.id,)
        )

    @staticmethod
    def clear(database: sqlite3.Connection) -> None:
        """
            Delete all Symbols from the symbol table.
            :param database: A database handle
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM symbol;'''
        )

    @staticmethod
    def get(database: sqlite3.Connection, elem_id: int) -> Symbol:
        """
            Return a symbol from the database with the matching id 
            :param database: A database handle
            :param elem_id: The id of the symbol to retrieve  
            :return: A Symbol object that reflect the content inside 
            the database 
        """
        out = SqliteHelper.fetch(database, '''
            SELECT * FROM symbol WHERE id = ?;''', (elem_id,)
        )

        if len(out) == 1:
            id_, type_ = out[0]
            return Symbol(id_, SymbolType(type_))

    @staticmethod
    def update(database: sqlite3.Connection, obj: Symbol) -> None:
        """
            Update a Symbol inside the symbol table.
            :param database: A database handle
            :param obj: The Symbol object to update 
            :return: None 
        """
        SqliteHelper.exec(database, '''
            UPDATE symbol SET
                definition_kind = ?
            WHERE
                id = ?;''', (obj.definition_kind.value, obj.id)
        )

    @staticmethod
    def list(database: sqlite3.Connection) -> list[Symbol]:
        """
            Return the list of all the symbols from the symbol table. 
            :param database: A database handle
            :return: The list of Symbols 
        """
        rows = SqliteHelper.fetch(database, '''
            SELECT * FROM symbol;'''
        )

        result = list()
        for row in rows:
            id_, type_ = row
            result.append(Symbol(id_, SymbolType(type_)))

        return result


class FileDAO(object):
    """
        This class is a static class that can manipulate File objects,
        inserting and removing them from a sqlite database.
    """

    @staticmethod
    def create_table(database: sqlite3.Connection) -> None:
        """
            Create the file table of the Sourcetrail database
            if it doesn't exist.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''    
            CREATE TABLE file(
                id INTEGER NOT NULL, 
                path TEXT, 
                language TEXT, 
                modification_time TEXT, 
                indexed INTEGER, 
                complete INTEGER, 
                line_count INTEGER, 
                PRIMARY KEY(id), 
                FOREIGN KEY(id) REFERENCES node(id) ON DELETE CASCADE
            );'''
        )

    @staticmethod
    def delete_table(database: sqlite3.Connection) -> None:
        """
            Delete the file table of the Sourcetrail database
            only if it exists.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''
            DROP TABLE IF EXISTS main.file;'''
        )

    @staticmethod
    def new(database: sqlite3.Connection, obj: File) -> int:
        """
            Insert a new File inside the file table.
            :param database: A database handle
            :param obj: The object to insert
            :return: The id of the inserted file
        """
        return SqliteHelper.exec(database, '''
            INSERT INTO file(
                id, path, language, modification_time, indexed, complete, line_count 
            ) VALUES(?, ?, ?, ?, ?, ?, ?);''', (
                obj.id, obj.path, obj.language, obj.modification_time,
                obj.indexed, obj.complete, obj.line_count
            )
        )

    @staticmethod
    def delete(database: sqlite3.Connection, obj: File) -> None:
        """
            Delete a File from the file table.
            :param database: A database handle
            :param obj: The object to delete
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM file WHERE id = ?;''', (obj.id,)
        )

    @staticmethod
    def clear(database: sqlite3.Connection) -> None:
        """
            Delete all Files from the file table.
            :param database: A database handle
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM file;'''
        )

    @staticmethod
    def get(database: sqlite3.Connection, elem_id: int) -> File:
        """
            Return a file from the database with the matching id 
            :param database: A database handle
            :param elem_id: The id of the file to retrieve  
            :return: A File object that reflect the content inside 
            the database 
        """
        out = SqliteHelper.fetch(database, '''
            SELECT * FROM file WHERE id = ?;''', (elem_id,)
        )

        if len(out) == 1:
            return File(*out[0])

    @staticmethod
    def update(database: sqlite3.Connection, obj: File) -> None:
        """
            Update a File inside the file table.
            :param database: A database handle
            :param obj: The File object to update 
            :return: None 
        """
        SqliteHelper.exec(database, '''
            UPDATE file SET
                path = ?,
                language = ?,
                modification_time = ?,
                indexed = ?,
                complete = ?,
                line_count = ? 
            WHERE
                id = ?;''', (
                obj.path, obj.language, obj.modification_time,
                obj.indexed, obj.complete, obj.line_count, obj.id
            )
        )

    @staticmethod
    def list(database: sqlite3.Connection) -> list[File]:
        """
            Return the list of all the files from the file table. 
            :param database: A database handle
            :return: The list of Files 
        """
        rows = SqliteHelper.fetch(database, '''
            SELECT * FROM file;'''
        )

        result = list()
        for row in rows:
            result.append(File(*row))

        return result


class FileContentDAO(object):
    """
        This class is a static class that can manipulate FileContent objects,
        inserting and removing them from a sqlite database.
    """

    @staticmethod
    def create_table(database: sqlite3.Connection) -> None:
        """
            Create the filecontent table of the Sourcetrail database
            if it doesn't exist.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''    
            CREATE TABLE filecontent(
                id INTEGER, 
                content TEXT, 
                PRIMARY KEY(id), 
                FOREIGN KEY(id) REFERENCES file(id)ON DELETE CASCADE ON UPDATE CASCADE
            );'''
        )

    @staticmethod
    def delete_table(database: sqlite3.Connection) -> None:
        """
            Delete the filecontent table of the Sourcetrail database
            only if it exists.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''
            DROP TABLE IF EXISTS main.filecontent;'''
        )

    @staticmethod
    def new(database: sqlite3.Connection, obj: FileContent) -> int:
        """
            Insert a new FileContent inside the filecontent table.
            :param database: A database handle
            :param obj: The object to insert
            :return: The id of the inserted filecontent
        """
        return SqliteHelper.exec(database, '''
            INSERT INTO filecontent(
                id, content 
            ) VALUES(?, ?);''', (obj.id, obj.content)
        )

    @staticmethod
    def delete(database: sqlite3.Connection, obj: FileContent) -> None:
        """
            Delete an FileContent from the filecontent table.
            :param database: A database handle
            :param obj: The object to delete
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM filecontent WHERE id = ?;''', (obj.id,)
        )

    @staticmethod
    def clear(database: sqlite3.Connection) -> None:
        """
            Delete all FileContents from the filecontent table.
            :param database: A database handle
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM filecontent;'''
        )

    @staticmethod
    def get(database: sqlite3.Connection, elem_id: int) -> FileContent:
        """
            Return a filecontent from the database with the matching id 
            :param database: A database handle
            :param elem_id: The id of the filecontent to retrieve  
            :return: A FileContent object that reflect the content inside 
            the database 
        """
        out = SqliteHelper.fetch(database, '''
            SELECT * FROM filecontent WHERE id = ?;''', (elem_id,)
        )

        if len(out) == 1:
            return FileContent(*out[0])

    @staticmethod
    def update(database: sqlite3.Connection, obj: FileContent) -> None:
        """
            Update an FileContent inside the filecontent table.
            :param database: A database handle
            :param obj: The FileContent object to update 
            :return: None 
        """
        SqliteHelper.exec(database, '''
            UPDATE filecontent SET
                content = ? 
            WHERE
                id = ?;''', (obj.content, obj.id)
        )

    @staticmethod
    def list(database: sqlite3.Connection) -> list[FileContent]:
        """
            Return the list of all the filecontents from the filecontent table. 
            :param database: A database handle
            :return: The list of FileContents 
        """
        rows = SqliteHelper.fetch(database, '''
            SELECT * FROM filecontent;'''
        )

        result = list()
        for row in rows:
            result.append(FileContent(*row))

        return result


class LocalSymbolDAO(object):
    """
        This class is a static class that can manipulate LocalSymbol objects,
        inserting and removing them from a sqlite database.
    """

    @staticmethod
    def create_table(database: sqlite3.Connection) -> None:
        """
            Create the local_symbol table of the Sourcetrail database
            if it doesn't exist.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''    
            CREATE TABLE local_symbol(
                id INTEGER NOT NULL, 
                name TEXT, 
                PRIMARY KEY(id), 
                FOREIGN KEY(id) REFERENCES element(id) ON DELETE CASCADE
            );'''
        )

    @staticmethod
    def delete_table(database: sqlite3.Connection) -> None:
        """
            Delete the local_symbol table of the Sourcetrail database
            only if it exists.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''
            DROP TABLE IF EXISTS main.local_symbol;'''
        )

    @staticmethod
    def new(database: sqlite3.Connection, obj: LocalSymbol) -> int:
        """
            Insert a new LocalSymbol inside the local_symbol table.
            :param database: A database handle
            :param obj: The object to insert
            :return: The id of the inserted local_symbol
        """
        return SqliteHelper.exec(database, '''
            INSERT INTO local_symbol(
                id, name 
            ) VALUES(?, ?);''', (obj.id, obj.name)
        )

    @staticmethod
    def delete(database: sqlite3.Connection, obj: LocalSymbol) -> None:
        """
            Delete an LocalSymbol from the local_symbol table.
            :param database: A database handle
            :param obj: The object to delete
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM local_symbol WHERE id = ?;''', (obj.id,)
        )

    @staticmethod
    def clear(database: sqlite3.Connection) -> None:
        """
            Delete all LocalSymbols from the local_symbol table.
            :param database: A database handle
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM local_symbol;'''
        )

    @staticmethod
    def get(database: sqlite3.Connection, elem_id: int) -> LocalSymbol:
        """
            Return a local_symbol from the database with the matching id 
            :param database: A database handle
            :param elem_id: The id of the local_symbol to retrieve  
            :return: A LocalSymbol object that reflect the content inside 
            the database 
        """
        out = SqliteHelper.fetch(database, '''
            SELECT * FROM local_symbol WHERE id = ?;''', (elem_id,)
        )

        if len(out) == 1:
            return LocalSymbol(*out[0])

    @staticmethod
    def get_from_name(database: sqlite3.Connection, name: str) -> LocalSymbol:
        """
            Return a local_symbol from the database with the matching name
            :param database: A database handle
            :param name: The name of the local_symbol to retrieve  
            :return: A LocalSymbol object that reflect the content inside 
            the database 
        """
        out = SqliteHelper.fetch(database, '''
            SELECT * FROM local_symbol WHERE name = ? LIMIT 1;''', (name,)
        )

        if len(out) == 1:
            return LocalSymbol(*out[0])

    @staticmethod
    def update(database: sqlite3.Connection, obj: LocalSymbol) -> None:
        """
            Update an LocalSymbol inside the local_symbol table.
            :param database: A database handle
            :param obj: The LocalSymbol object to update 
            :return: None 
        """
        SqliteHelper.exec(database, '''
            UPDATE local_symbol SET
                name = ? 
            WHERE
                id = ?;''', (obj.name, obj.id)
        )

    @staticmethod
    def list(database: sqlite3.Connection) -> list[LocalSymbol]:
        """
            Return the list of all the local_symbols from the local_symbol table. 
            :param database: A database handle
            :return: The list of LocalSymbols 
        """
        rows = SqliteHelper.fetch(database, '''
            SELECT * FROM local_symbol;'''
        )

        result = list()
        for row in rows:
            result.append(LocalSymbol(*row))

        return result


class SourceLocationDAO(object):
    """
        This class is a static class that can manipulate SourceLocation objects,
        inserting and removing them from a sqlite database.
    """

    @staticmethod
    def create_table(database: sqlite3.Connection) -> None:
        """
            Create the source_location table of the Sourcetrail database
            if it doesn't exist.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''    
            CREATE TABLE source_location(
                id INTEGER NOT NULL, 
                file_node_id INTEGER, 
                start_line INTEGER, 
                start_column INTEGER, 
                end_line INTEGER, 
                end_column INTEGER, 
                type INTEGER, 
                PRIMARY KEY(id), 
                FOREIGN KEY(file_node_id) REFERENCES node(id) ON DELETE CASCADE
            );'''
        )

    @staticmethod
    def delete_table(database: sqlite3.Connection) -> None:
        """
            Delete the source_location table of the Sourcetrail database
            only if it exists.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''
            DROP TABLE IF EXISTS main.source_location;'''
        )

    @staticmethod
    def new(database: sqlite3.Connection, obj: SourceLocation) -> int:
        """
            Insert a new SourceLocation inside the source_location table.
            :param database: A database handle
            :param obj: The object to insert
            :return: The id of the inserted source_location
        """
        return SqliteHelper.exec(database, '''
            INSERT INTO source_location(
                id, file_node_id, start_line, start_column, end_line, end_column, type 
            ) VALUES(NULL, ?, ?, ?, ?, ?, ?);''', (
                obj.file_node_id, obj.start_line, obj.start_column,
                obj.end_line, obj.end_column, obj.type.value
            )
        )

    @staticmethod
    def delete(database: sqlite3.Connection, obj: SourceLocation) -> None:
        """
            Delete an SourceLocation from the source_location table.
            :param database: A database handle
            :param obj: The object to delete
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM source_location WHERE id = ?;''', (obj.id,)
        )

    @staticmethod
    def clear(database: sqlite3.Connection) -> None:
        """
            Delete all SourceLocations from the source_location table.
            :param database: A database handle
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM source_location;'''
        )

    @staticmethod
    def get(database: sqlite3.Connection, elem_id: int) -> SourceLocation:
        """
            Return a source_location from the database with the matching id 
            :param database: A database handle
            :param elem_id: The id of the source_location to retrieve  
            :return: A SourceLocation object that reflect the content inside 
            the database 
        """
        out = SqliteHelper.fetch(database, '''
            SELECT * FROM source_location WHERE id = ?;''', (elem_id,)
        )

        if len(out) == 1:
            id_, fid, sl, sc, el, ec, type_ = out[0]
            return SourceLocation(id_, fid, sl, sc, el, ec,
                  SourceLocationType(type_)
            )

    @staticmethod
    def update(database: sqlite3.Connection, obj: SourceLocation) -> None:
        """
            Update an SourceLocation inside the source_location table.
            :param database: A database handle
            :param obj: The SourceLocation object to update 
            :return: None 
        """
        SqliteHelper.exec(database, '''
            UPDATE source_location SET
                file_node_id = ?,
                start_line = ?,
                start_column = ?,
                end_line = ?,
                end_column = ?,
                type = ? 
            WHERE
                id = ?;''', (
                obj.file_node_id, obj.start_line, obj.start_column,
                obj.end_line, obj.end_column, obj.type.value, obj.id
            )
        )

    @staticmethod
    def list(database: sqlite3.Connection) -> list[SourceLocation]:
        """
            Return the list of all the source_locations from the source_location table. 
            :param database: A database handle
            :return: The list of SourceLocations 
        """
        rows = SqliteHelper.fetch(database, '''
            SELECT * FROM source_location;'''
        )

        result = list()
        for row in rows:
            id_, fid, sl, sc, el, ec, type_ = row
            result.append(SourceLocation(
                id_, fid, sl, sc, el, ec, SourceLocationType(type_)
            ))

        return result


class OccurrenceDAO(object):
    """
        This class is a static class that can manipulate Occurrence objects,
        inserting and removing them from a sqlite database.
    """

    @staticmethod
    def create_table(database: sqlite3.Connection) -> None:
        """
            Create the occurrence table of the Sourcetrail database
            if it doesn't exist.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''    
            CREATE TABLE occurrence(
                element_id INTEGER NOT NULL, 
                source_location_id INTEGER NOT NULL, 
                PRIMARY KEY(element_id, source_location_id), 
                FOREIGN KEY(element_id) REFERENCES element(id) ON DELETE CASCADE, 
                FOREIGN KEY(source_location_id) REFERENCES source_location(id) 
                    ON DELETE CASCADE
            );'''
        )

    @staticmethod
    def delete_table(database: sqlite3.Connection) -> None:
        """
            Delete the occurrence table of the Sourcetrail database
            only if it exists.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''
            DROP TABLE IF EXISTS main.occurrence;'''
        )

    @staticmethod
    def new(database: sqlite3.Connection, obj: Occurrence) -> int:
        """
            Insert a new Occurrence inside the occurrence table.
            :param database: A database handle
            :param obj: The object to insert
            :return: The id of the inserted occurrence
        """
        return SqliteHelper.exec(database, '''
            INSERT INTO occurrence(
                element_id, source_location_id 
            ) VALUES(?, ?);''', (obj.element_id, obj.source_location_id)
        )

    @staticmethod
    def delete(database: sqlite3.Connection, obj: Occurrence) -> None:
        """
            Delete an Occurrence from the occurrence table.
            :param database: A database handle
            :param obj: The object to delete
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM occurrence WHERE element_id = ?;''', (obj.element_id,)
        )

    @staticmethod
    def clear(database: sqlite3.Connection) -> None:
        """
            Delete all Occurrences from the occurrence table.
            :param database: A database handle
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM occurrence;'''
        )

    @staticmethod
    def get(database: sqlite3.Connection, elem_id: int) -> Occurrence:
        """
            Return an occurrence from the database with the matching id
            :param database: A database handle
            :param elem_id: The id of the occurrence to retrieve  
            :return: A Occurrence object that reflect the content inside 
            the database 
        """
        out = SqliteHelper.fetch(database, '''
            SELECT * FROM occurrence WHERE element_id = ?;''', (elem_id,)
        )

        if len(out) == 1:
            return Occurrence(*out[0])

    @staticmethod
    def update(database: sqlite3.Connection, obj: Occurrence) -> None:
        """
            Update an Occurrence inside the occurrence table.
            :param database: A database handle
            :param obj: The Occurrence object to update 
            :return: None 
        """
        SqliteHelper.exec(database, '''
            UPDATE occurrence SET
                source_location_id = ? 
            WHERE
                element_id = ?;''', (obj.source_location_id, obj.element_id)
        )

    @staticmethod
    def list(database: sqlite3.Connection) -> list[Occurrence]:
        """
            Return the list of all the occurrences from the occurrence table. 
            :param database: A database handle
            :return: The list of Occurrences 
        """
        rows = SqliteHelper.fetch(database, '''
            SELECT * FROM occurrence;'''
        )

        result = list()
        for row in rows:
            result.append(Occurrence(*row))

        return result


class ComponentAccessDAO(object):
    """
        This class is a static class that can manipulate ComponentAccess objects,
        inserting and removing them from a sqlite database.
    """

    @staticmethod
    def create_table(database: sqlite3.Connection) -> None:
        """
            Create the component_access table of the Sourcetrail database
            if it doesn't exist.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''    
            CREATE TABLE component_access(
                node_id INTEGER NOT NULL, 
                type INTEGER NOT NULL, 
                PRIMARY KEY(node_id), 
                FOREIGN KEY(node_id) REFERENCES node(id) ON DELETE CASCADE
            );'''
        )

    @staticmethod
    def delete_table(database: sqlite3.Connection) -> None:
        """
            Delete the component_access table of the Sourcetrail database
            only if it exists.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''
            DROP TABLE IF EXISTS main.component_access;'''
        )

    @staticmethod
    def new(database: sqlite3.Connection, obj: ComponentAccess) -> int:
        """
            Insert a new ComponentAccess inside the component_access table.
            :param database: A database handle
            :param obj: The object to insert
            :return: The id of the inserted component_access
        """
        return SqliteHelper.exec(database, '''
            INSERT INTO component_access(
                node_id, type 
            ) VALUES(?, ?);''', (obj.node_id, obj.type.value)
        )

    @staticmethod
    def delete(database: sqlite3.Connection, obj: ComponentAccess) -> None:
        """
            Delete an ComponentAccess from the component_access table.
            :param database: A database handle
            :param obj: The object to delete
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM component_access WHERE node_id = ?;''', (obj.node_id,)
        )

    @staticmethod
    def clear(database: sqlite3.Connection) -> None:
        """
            Delete all ComponentAccess from the component_access table.
            :param database: A database handle
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM component_access;'''
        )

    @staticmethod
    def get(database: sqlite3.Connection, elem_id: int) -> ComponentAccess:
        """
            Return a component_access from the database with the matching id 
            :param database: A database handle
            :param elem_id: The id of the component_access to retrieve  
            :return: A ComponentAccess object that reflect the content inside 
            the database 
        """
        out = SqliteHelper.fetch(database, '''
            SELECT * FROM component_access WHERE node_id = ?;''', (elem_id,)
        )

        if len(out) == 1:
            node_id, type_ = out[0]
            return ComponentAccess(node_id, ComponentAccessType(type_))

    @staticmethod
    def update(database: sqlite3.Connection, obj: ComponentAccess) -> None:
        """
            Update an ComponentAccess inside the component_access table.
            :param database: A database handle
            :param obj: The ComponentAccess object to update 
            :return: None 
        """
        SqliteHelper.exec(database, '''
            UPDATE component_access SET
                type = ? 
            WHERE
                node_id = ?;''', (obj.type.value, obj.node_id)
        )

    @staticmethod
    def list(database: sqlite3.Connection) -> list[ComponentAccess]:
        """
            Return the list of all the component_access from the component_access table.
            :param database: A database handle
            :return: The list of ComponentAccess
        """
        rows = SqliteHelper.fetch(database, '''
            SELECT * FROM component_access;'''
        )

        result = list()
        for row in rows:
            node_id, type_ = row
            result.append(ComponentAccess(node_id, ComponentAccessType(type_)))

        return result


class ErrorDAO(object):
    """
        This class is a static class that can manipulate Error objects,
        inserting and removing them from a sqlite database.
    """

    @staticmethod
    def create_table(database: sqlite3.Connection) -> None:
        """
            Create the error table of the Sourcetrail database
            if it doesn't exist.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''    
            CREATE TABLE error(
                id INTEGER NOT NULL, 
                message TEXT, 
                fatal INTEGER NOT NULL, 
                indexed INTEGER NOT NULL, 
                translation_unit TEXT, 
                PRIMARY KEY(id), 
                FOREIGN KEY(id) REFERENCES element(id) ON DELETE CASCADE
            );'''
        )

    @staticmethod
    def delete_table(database: sqlite3.Connection) -> None:
        """
            Delete the error table of the Sourcetrail database
            only if it exists.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''
            DROP TABLE IF EXISTS main.error;'''
        )

    @staticmethod
    def new(database: sqlite3.Connection, obj: Error) -> int:
        """
            Insert a new Error inside the error table.
            :param database: A database handle
            :param obj: The object to insert
            :return: The id of the inserted error
        """
        return SqliteHelper.exec(database, '''
            INSERT INTO error(
                id, message, fatal, indexed, translation_unit 
            ) VALUES(?, ?, ?, ?, ?);''', (
            obj.id, obj.message, obj.fatal, obj.indexed, obj.translation_unit
        )
                                 )

    @staticmethod
    def delete(database: sqlite3.Connection, obj: Error) -> None:
        """
            Delete an Error from the error table.
            :param database: A database handle
            :param obj: The object to delete
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM error WHERE id = ?;''', (obj.id,)
        )

    @staticmethod
    def get(database: sqlite3.Connection, elem_id: int) -> Error:
        """
            Return an error from the database with the matching id
            :param database: A database handle
            :param elem_id: The id of the error to retrieve  
            :return: A Error object that reflect the content inside 
            the database 
        """
        out = SqliteHelper.fetch(database, '''
            SELECT * FROM error WHERE id = ?;''', (elem_id,)
        )

        if len(out) == 1:
            return Error(*out[0])

    @staticmethod
    def clear(database: sqlite3.Connection) -> None:
        """
            Delete all Errors from the error table.
            :param database: A database handle
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM error;'''
        )

    @staticmethod
    def update(database: sqlite3.Connection, obj: Error) -> None:
        """
            Update an Error inside the error table.
            :param database: A database handle
            :param obj: The Error object to update 
            :return: None 
        """
        SqliteHelper.exec(database, '''
            UPDATE error SET
                message = ?,
                fatal = ?,
                indexed = ?,
                translation_unit = ?
            WHERE
                id = ?;''', (
                obj.message, obj.fatal, obj.indexed,
                obj.translation_unit, obj.id
            )
        )

    @staticmethod
    def list(database: sqlite3.Connection) -> list[Error]:
        """
            Return the list of all the errors from the error table. 
            :param database: A database handle
            :return: The list of Errors 
        """
        rows = SqliteHelper.fetch(database, '''
            SELECT * FROM error;'''
        )

        result = list()
        for row in rows:
            result.append(Error(*row))

        return result


class MetaDAO(object):
    """
        This class is a static class that can manipulate Meta information,
        inserting and removing them from a sqlite database. There is no
        Meta object but a simple key, value pair can be used. 
    """

    @staticmethod
    def create_table(database: sqlite3.Connection) -> None:
        """
            Create the meta table of the Sourcetrail database
            if it doesn't exist.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''    
            CREATE TABLE meta(
                id INTEGER, 
                key TEXT, 
                value TEXT, 
                PRIMARY KEY(id)
            );'''
        )

    @staticmethod
    def delete_table(database: sqlite3.Connection) -> None:
        """
            Delete the meta table of the Sourcetrail database
            only if it exists.
            :param database: A database handle
            :return: None
        """
        SqliteHelper.exec(database, '''
            DROP TABLE IF EXISTS main.meta;'''
        )

    @staticmethod
    def new(database: sqlite3.Connection, key: str, value: str) -> int:
        """
            Insert a new Meta inside the meta table.
            :param database: A database handle
            :param key: The key to insert 
            :param value: The value to insert
            :return: The id of the inserted meta
        """
        return SqliteHelper.exec(database, '''
            INSERT INTO meta(
                id, key, value  
            ) VALUES(NULL, ?, ?);''', (
                key, value
            )
        )

    @staticmethod
    def delete(database: sqlite3.Connection, id_: int) -> None:
        """
            Delete a Meta from the meta table.
            :param database: A database handle
            :param id_: The identifier of the object to delete
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM meta WHERE id = ?;''', (id_,)
        )

    @staticmethod
    def get(database: sqlite3.Connection, id_: int) -> tuple[int, str, str]:
        """
            Return a meta from the database with the matching id 
            :param database: A database handle
            :param id_: The id of the meta to retrieve  
            :return: A Meta object that reflect the content inside 
            the database 
        """
        out = SqliteHelper.fetch(database, '''
            SELECT id, key, value FROM meta WHERE id = ?;''', (id_,)
        )
        return tuple(out[0])

    @staticmethod
    def clear(database: sqlite3.Connection) -> None:
        """
            Delete all Metas from the meta table.
            :param database: A database handle
            :return: None 
        """
        SqliteHelper.exec(database, '''
            DELETE FROM meta;'''
        )

    @staticmethod
    def update(database: sqlite3.Connection, id_: int, key: str,
               value: str) -> None:
        """
            Update a Meta inside the meta table.
            :param database: A database handle
            :param id_: The id of the meta to update  
            :param key: The key to update  
            :param value: The value to insert
            :return: None 
        """
        SqliteHelper.exec(database, '''
            UPDATE meta SET
                key = ?,
                value = ?
            WHERE
                id = ?;''', (
                key, value, id_
            )
        )

    @staticmethod
    def list(database: sqlite3.Connection) -> list[tuple[int, str, str]]:
        """
            Return the list of all the metas from the meta table. 
            :param database: A database handle
            :return: The list of Metas 
        """
        rows = SqliteHelper.fetch(database, '''
            SELECT * FROM meta;'''
        )

        result = list()
        for row in rows:
            result.append(tuple(row))

        return result



# -*- coding: utf-8 -*-

#  Copyright 2023 Quarkslab
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import enum

from .exceptions import SerializeException, DeserializeException


# ------------------------------------------------------------------------ #
# Basic wrapper Types for API                                              #
# ------------------------------------------------------------------------ #

class Element(object):
    """
        Wrapper class for sourcetrail 'element' table:

        ```sql
        CREATE TABLE element(
            id INTEGER, 
            PRIMARY KEY(id)
        )
        ```

        The 'element' table is used in sourcetrail to be able to easily manage
        other elements of others tables. Since all higher level element are
        referencing an element in the 'element' table it's possible to remove
        any element by removing the correct entry in the 'element' table.
    """

    def __init__(self, id_: int = 0):
        """
            Create a new Element object. 
            :param id_: The id of the element
        """

        self.id = id_


class ElementComponentType(enum.Enum):
    """
        Internal class that represent an ElementComponent type 
        inside the sourcetrail database. This type is used to 
        indicate that an element component is ambiguous.  
    """

    NONE = 0
    IS_AMBIGUOUS = 1


class ElementComponent(Element):
    """ 
        Wrapper class for sourcetrail element_component table:

        ```sql
        CREATE TABLE element_component(
            id INTEGER, 	
            element_id INTEGER, 	
            type INTEGER, 	
            data TEXT, 	
            PRIMARY KEY(id), 	
            FOREIGN KEY(element_id) REFERENCES element(id) ON DELETE CASCADE
        )
        ```

        This table is not commonly used, it only contains indication about the
        ambiguity of another element such as an edge or a node.
    """

    def __init__(self, id_: int = 0, elem_id: int = 0,
                 type_: ElementComponentType = ElementComponentType.NONE,
                 data: str = ''):
        """
            Create a new ElementComponent object. 
            :param id_: The id of the element 
            :param elem_id: The id of the referenced element
            :param type_: The type of the ElementComponent 
            :param data: Additional data (optional)
        """

        super().__init__(id_)
        self.elem_id = elem_id
        self.type = type_
        self.data = data


class EdgeType(enum.Enum):
    """
        Internal class that represent an Edge type inside the
        sourcetrail database. This type define the relationship
        between the nodes. 
    """
    UNDEFINED = 0
    MEMBER = 1 << 0
    TYPE_USAGE = 1 << 1
    USAGE = 1 << 2
    CALL = 1 << 3
    INHERITANCE = 1 << 4
    OVERRIDE = 1 << 5
    TYPE_ARGUMENT = 1 << 6
    TEMPLATE_SPECIALIZATION = 1 << 7
    INCLUDE = 1 << 8
    IMPORT = 1 << 9
    BUNDLED_EDGES = 1 << 10
    MACRO_USAGE = 1 << 11
    ANNOTATION_USAGE = 1 << 12


class Edge(Element):
    """ 
        Wrapper class for sourcetrail edge table:

        ```sql
        CREATE TABLE edge(
            id INTEGER NOT NULL, 
            type INTEGER NOT NULL, 
            source_node_id INTEGER NOT NULL, 
            target_node_id INTEGER NOT NULL, 
            PRIMARY KEY(id), 
            FOREIGN KEY(id) REFERENCES element(id) ON DELETE CASCADE, 
            FOREIGN KEY(source_node_id) REFERENCES node(id) ON DELETE CASCADE, 
            FOREIGN KEY(target_node_id) REFERENCES node(id) ON DELETE CASCADE
        ) 
        ```

        The 'edge' table is used to define relation between element of the 'node'
        table. For example, it can be used to indicate that a field is a member
        of a class or a function foo is calling another function bar.
    """

    def __init__(self, id_: int = 0, type_: EdgeType = EdgeType.UNDEFINED,
                 src: int = 0, dst: int = 0):
        """
            Create a new Edge object. 
            :param id_: The id of the element 
            :param type_: The type of the Edge 
            :param src: The id of the source element 
            :param dst: The id of the destination element
        """
 
        super().__init__(id_)
        self.type = type_
        self.src = src
        self.dst = dst


class NodeType(enum.Enum):
    """
        Internal class that represent an Edge type inside the
        sourcetrail database. This type define the type of the
        node in the database. 
    """
    NODE_SYMBOL = 1 << 0
    NODE_TYPE = 1 << 1
    NODE_BUILTIN_TYPE = 1 << 2
    NODE_MODULE = 1 << 3
    NODE_NAMESPACE = 1 << 4
    NODE_PACKAGE = 1 << 5
    NODE_STRUCT = 1 << 6
    NODE_CLASS = 1 << 7
    NODE_INTERFACE = 1 << 8
    NODE_ANNOTATION = 1 << 9
    NODE_GLOBAL_VARIABLE = 1 << 10
    NODE_FIELD = 1 << 11
    NODE_FUNCTION = 1 << 12
    NODE_METHOD = 1 << 13
    NODE_ENUM = 1 << 14
    NODE_ENUM_CONSTANT = 1 << 15
    NODE_TYPEDEF = 1 << 16
    NODE_TYPE_PARAMETER = 1 << 17
    NODE_FILE = 1 << 18
    NODE_MACRO = 1 << 19
    NODE_UNION = 1 << 20


class Node(Element):
    """
        Wrapper class for sourcetrail node table:

        ```sql
        CREATE TABLE node(
            id INTEGER NOT NULL, 
            type INTEGER NOT NULL, 
            serialized_name TEXT, 
            PRIMARY KEY(id), 
            FOREIGN KEY(id) REFERENCES element(id) ON DELETE CASCADE
        )
        ```

        The 'node' table is the main table of the sourcetrail database. It allows
        to store elements such as function, class, package, etc. However, this
        table is weirdly implemented as the field 'serialized_name' contains 
        another type called NameHierarchy with a custom serialization format.

        The NameHierarchy describe a relationship between node elements, for
        example, a class 'A' with a member 'b' will result in two entries in the
        database:

            id1 | NODE_CLASS | '.\tA\ts\tp' 
            id2 | NODE_FIELD | '.\tA\ts\tp\tm\tb\ts\tp'

        The above example shows that the 'serialized_name' of the member 'b' (id2)
        hold some information about his parent 'A' (id1).
    """

    def __init__(self, id_: int = 0, type_: NodeType = NodeType.NODE_TYPE,
                 name: str = ''):
        """
            Create a new Node object. 
            :param id_: The id of the element 
            :param type_: The type of the Node 
            :param name: The serialized name of the Node
        """
 
        super().__init__(id_)
        self.type = type_
        self.name = name


class SymbolType(enum.Enum):
    """
        Internal class that represent a Symbol type inside the 
        sourcetrail database. This type define the symbol type in the database.
    """
    NONE = 0
    IMPLICIT = 1
    EXPLICIT = 2


class Symbol(Element):
    """
        Wrapper class for sourcetrail symbol table:

        ```sql
        CREATE TABLE symbol(
            id INTEGER NOT NULL, 
            definition_kind INTEGER NOT NULL, 
            PRIMARY KEY(id), 
            FOREIGN KEY(id) REFERENCES node(id) ON DELETE CASCADE
        ) 
        ```

        The 'symbol' table is used to add additional information on elements 
        such as node.  
    """

    def __init__(self, id_: int = 0, definition: SymbolType = SymbolType.NONE):
        """
            Create a new Symbol object. 
            :param id_: The id of the element 
            :param definition: The type of the Symbol
        """
  
        super().__init__(id_)
        self.definition_kind = definition


class File(Element):
    """
        Wrapper class for sourcetrail file table:

        ```sql
        CREATE TABLE file(
            id INTEGER NOT NULL, 
            path TEXT, 
            language TEXT, 
            modification_time TEXT, 
            indexed INTEGER, 
            complete INTEGER, 
            line_count INTEGER, 
            PRIMARY KEY(id), 
            FOREIGN KEY(id) REFERENCES node(id) ON DELETE CASCADE
        )
        ```

        The 'file' table hold the information about the different source file
        that have been parsed by sourcetrail. Each one of them contains an id that 
        will be reference by the element of the 'filecontent' table.
    """

    def __init__(self, id_: int = 0, path: str = '', language: str = '',
                 modification_time: str = '', indexed: int = 0, complete: int = 0,
                 line_count: int = 0):
        """
            Create a new File object.

            :param id_: The id of the element 
            :param path: The path to the source file 
            :param language: The language of the source file.
            :param modification_time: The time of the last modification of the source file
            :param indexed: A indication to tell if the file was indexed or not (0 or 1)
            :param complete: A indicate to tell if the indexing is complete or not (0 or 1)
            :param line_count: The number of line in the source file
        """
 
        super().__init__(id_)
        self.path = path
        self.language = language
        self.modification_time = modification_time
        self.indexed = indexed
        self.complete = complete
        self.line_count = line_count


class FileContent(Element):
    """
        Wrapper class for sourcetrail filecontent table:

        ```sql
        CREATE TABLE filecontent(
            id INTEGER, 
            content TEXT, 
            PRIMARY KEY(id), 
            FOREIGN KEY(id) REFERENCES file(id)ON DELETE CASCADE ON UPDATE CASCADE
        )
        ```

        The 'filecontent' table holds the content of the different source file. 
        Because the id field of the filecontent is a primary key (unique element)
        a filecontent should contain the entire content of a file.
    """

    def __init__(self, id_: int = 0, content: str = ''):
        """
            Create a new FileContent object.

            :param id_: The id of the element 
            :param content: The content of the source file.
        """
 
        super().__init__(id_)
        self.content = content


class LocalSymbol(Element):
    """
        Wrapper class for sourcetrail local_symbol table:

        ```sql
        CREATE TABLE local_symbol(
            id INTEGER NOT NULL, 
            name TEXT, 
            PRIMARY KEY(id), 
            FOREIGN KEY(id) REFERENCES element(id) ON DELETE CASCADE
        )
        ```
        
        The 'local_symbol' table holds reference to nodes that represent
        elements such as variables only used locally. For example, the
        following code snippet:

        ```c
        int foo(int a)
        {
            my_global_func();
            int b = 2 * a;
            return b;
        }
        ```

        While result in 3 different nodes, two of them (a and b) with have
        an entry in the 'local_symbol' table. The content of this table is
        not essential to the application but can provide a better level of
        detail to the user when browsing code in Sourcetrail UI.
    """

    def __init__(self, id_: int = 0, name: str = ''):
        """
            Create a new LocalSymbol object.

            :param id_: The id of the element 
            :param name: The name of local symbol
        """
 
        super().__init__(id_)
        self.name = name


class SourceLocationType(enum.Enum):
    """
        Internal class that represent a SymbolLocation type inside the 
        sourcetrail database. This type define the symbol location type in the
        database and can be used to indicate indexing errors.
    """
    TOKEN = 0
    SCOPE = 1
    QUALIFIER = 2
    LOCAL_SYMBOL = 3
    SIGNATURE = 4
    ATOMIC_RANGE = 5
    INDEXER_ERROR = 6
    FULLTEXT_SEARCH = 7
    SCREEN_SEARCH = 8
    UNSOLVED = 9


class SourceLocation(Element):
    """
        Wrapper class for sourcetrail source_location table:

        ```sql
        CREATE TABLE source_location(
            id INTEGER NOT NULL, 
            file_node_id INTEGER, 
            start_line INTEGER, 
            start_column INTEGER, 
            end_line INTEGER, 
            end_column INTEGER, 
            type INTEGER, 
            PRIMARY KEY(id), 
            FOREIGN KEY(file_node_id) REFERENCES node(id) ON DELETE CASCADE
        )
        ```

        The table 'source_location' contain some entries indicating where the
        corresponding node element (the ones that are referenced by the id) are
        located in the source tree.
    """

    def __init__(self, id_: int = 0, file_node_id: int = 0, start_line: int = 0,
                 start_column: int = 0, end_line: int = 0, end_column: int = 0,
                 type_: SourceLocationType = SourceLocationType.UNSOLVED):
        """
            Create a new SourceLocation object.

            :param id_: The id of the element 
            :param file_node_id: The id of the file element corresponding to this content. 
            :param start_line: The line at which the element starts.
            :param start_column: The column at which the element starts.
            :param end_line: The line at which the element ends.
            :param end_column: The line at which the element ends.
            :param type_: The type of the source location.
        """
 
        super().__init__(id_)
        self.file_node_id = file_node_id
        self.start_line = start_line
        self.start_column = start_column
        self.end_line = end_line
        self.end_column = end_column
        self.type = type_


class Occurrence(object):
    """
        Wrapper class for sourcetrail occurrence table:

        ```sql
        CREATE TABLE occurrence(
            element_id INTEGER NOT NULL, 
            source_location_id INTEGER NOT NULL, 
            PRIMARY KEY(element_id, source_location_id), 
            FOREIGN KEY(element_id) REFERENCES element(id) ON DELETE CASCADE, 
            FOREIGN KEY(source_location_id) REFERENCES source_location(id) 
                ON DELETE CASCADE
        )
        ```

        The table 'occurrence' allow to link the elements define in the 'source_location'
        table and the ones define in the 'node' table. 
    """

    def __init__(self, elem_id: int = 0, source_location_id: int = 0):
        """
            Create a new Occurrence object.

            :param elem_id: The id of the element referenced by this occurrence
            :param source_location_id: The id of the source location referenced by this occurrence
        """
 
        self.element_id = elem_id
        self.source_location_id = source_location_id


class ComponentAccessType(enum.Enum):
    """
        Internal class that represent a ComponentAccess type inside 
        the sourcetrail database. This type define the type of 
        ComponentAccess in the database and can be used to add extra
        information regarding the visibility of elements (in the 
        "Object-Oriented" sens)
    """
    NONE = 0
    PUBLIC = 1
    PROTECTED = 2
    PRIVATE = 3
    DEFAULT = 4
    TEMPLATE_PARAMETER = 5
    TYPE_PARAMETER = 6


class ComponentAccess(object):
    """
        Wrapper class for sourcetrail component_access table:

        ```sql
        CREATE TABLE component_access(
            node_id INTEGER NOT NULL, 
            type INTEGER NOT NULL, 
            PRIMARY KEY(node_id), 
            FOREIGN KEY(node_id) REFERENCES node(id) ON DELETE CASCADE
        )
        ```

        The table 'component_access' allow to add information on the type
        of visibility of element in the 'node' table. For example, a java
        class that is set to public will have an entry in this table.
    """

    def __init__(self, node_id: int = 0,
                 type_: ComponentAccessType = ComponentAccessType.NONE):
        """
            Create a new ComponentAccess object.

            :param node_id: The id of the element 
            :param type_: The type of the ComponentAccess
        """
 
        self.node_id = node_id
        self.type = type_


class Error(Element):
    """
        Wrapper class for sourcetrail error table:

        ```sql
        CREATE TABLE error(
            id INTEGER NOT NULL, 
            message TEXT, 
            fatal INTEGER NOT NULL, 
            indexed INTEGER NOT NULL, 
            translation_unit TEXT, 
            PRIMARY KEY(id), 
            FOREIGN KEY(id) REFERENCES element(id) ON DELETE CASCADE
        )
        ```

        The table 'error' holds the different error message produced during
        the parsing of the source files and are displayed in the UI. 
    """

    def __init__(self, id_: int = 0, message: str = '', fatal: int = 0,
                 indexed: int = 0, translation_unit: str = ''):
        """
            Create a new Error object.

            :param id_: The id of the element 
            :param message: The description of the error 
            :param fatal: Indicate if this error is fatal or not (0 or 1)
            :param indexed: Indicate if this error occurs while indexing (0 or 1)
            :param translation_unit: Indicate in which translation unit the error occurs
        """
 
        super().__init__(id_)
        self.message = message
        self.fatal = fatal
        self.indexed = indexed
        self.translation_unit = translation_unit


class NameElement(object):
    """
        This class is a basic component of the serialized_name field
        of the node table. This type does not represent directly a 
        type in the database.
    """

    def __init__(self, prefix: str = None, name: str = None,
                 postfix: str = None):
        """
            Create a new NameElement object.

            :param prefix: The prefix of the element 
            :param name: The name of the element
            :param postfix: The postfix of the element
        """
 
        self._prefix = prefix
        self._name = name
        self._postfix = postfix

    def __str__(self) -> str:
        """
            Return a textual description of this NameElement

            :return: A description of this object 
        """

        return '<NameElement prefix:%s, name: %s, postfix: %s>' % (
            self._prefix, self._name, self._postfix
        )

    def get_prefix(self) -> str:
        """
            Return the prefix of this element

            :return: The prefix of the element
        """

        return self._prefix

    def get_name(self) -> str:
        """
            Return the name of this element

            :return: The name of the element
        """

        return self._name

    def get_postfix(self) -> str:
        """
            Return the postfix of this element

            :return: The postfix of the element
        """

        return self._postfix

    def set_prefix(self, prefix: str) -> None:
        """
            Set the prefix of this element

            :param prefix: The new prefix of the element 
            :return: None 
        """
 
        self._prefix = prefix

    def set_name(self, name: str) -> None:
        """
            Set the name of this element

            :param name: The new name of the element 
            :return: None 
        """

        self._name = name

    def set_postfix(self, postfix: str) -> None:
        """
            Set the postfix of this element

            :param postfix: The new postfix of the element 
            :return: None 
        """

        self._postfix = postfix


class NameHierarchy(object):
    """
        This class represent the hierarchy relationship between nodes. 
        This type does not represent directly a type in the database and
        is stored in the 'serialized_name' field of the 'node' table.
    """

    # Delimiters for the serialized_name
    DELIMITER = '\t'
    META_DELIMITER = '\tm'
    NAME_DELIMITER = '\tn'
    PART_DELIMITER = '\ts'
    SIGNATURE_DELIMITER = '\tp'

    # Name delimiter type
    NAME_DELIMITER_FILE = '/'
    NAME_DELIMITER_CXX = '::'
    NAME_DELIMITER_JAVA = '.'
    NAME_DELIMITER_UNKNOWN = '@'

    NAME_DELIMITERS = [
        NAME_DELIMITER_FILE,
        NAME_DELIMITER_CXX,
        NAME_DELIMITER_JAVA,
        NAME_DELIMITER_UNKNOWN
    ]

    def __init__(self, delimiter: str, elements: list[NameElement]):
        """
            Create a new NameHierarchy object

            :param delimiter: The delimiter of for this NameHierarchy, must be one of the NAME_DELIMITERS
            :param elements: A list of NameElement representing the hierarchy of this element
            :return: None
        """

        self._delimiter = delimiter
        self._elements = elements

    def __serialize(self, start: int, end: int) -> str:
        """
            Wrapper for serialize_range and serialize_name. 

            The serialization process is the following:
                - Start with any of the NAME_DELIMITERS
                - Serialize the **first element** by:
                    - Adding the name field concatenated with NAME_DELIMITER
                    - Add the prefix concatenated with PART_DELIMITER
                    - Add the postfix concatenated with SIGNATURE_DELIMITER
                - Remainders elements are concatenated in the same way that they
                  are separated with a META_DELIMITER

            :param start: the starting position of the elements  
            :param end: the ending position of the elements  
            :return: The serialized name 
        """
        result = self._delimiter + self.META_DELIMITER
        if not self._elements:
            # @Warning: This is not the same behavior as SourcetrailDB
            # We are returning an exception instead of an empty string
            raise SerializeException()

        elements = self._elements[start:end]
        for i, elem in enumerate(elements):
            if i > 0:
                result += self.NAME_DELIMITER

            result += elem.get_name() + self.PART_DELIMITER
            result += elem.get_prefix() + self.SIGNATURE_DELIMITER
            result += elem.get_postfix()
        return result

    def extend(self, element: NameElement) -> None:
        """
            Utility method that adds a new element to a hierarchy

            :param element: The new element to add
        """
        self._elements.append(element)

    def serialize_range(self, start: int, end: int) -> str:
        """
            Utility method that return a part of the serialized name

            :param start: the starting position of the elements
            :param end: the ending position of the elements  
            :return: The serialized name 
        """
        return self.__serialize(start, end)

    def serialize_name(self) -> str:
        """
            Utility method that return the full serialized name

            :return: The serialized name
        """
        return self.__serialize(0, self.size())

    def size(self) -> int:
        """
            Return the size of the NameHierarchy object.

            :return: The size of the NameHierarchy
        """
        return len(self._elements) if self._elements else 0

    @staticmethod
    def deserialize_name(serialized_name: str) -> 'NameHierarchy':
        """
            Utility method that return prefix, name and suffix 
            from a serialized name

            :param serialized_name: A string that should start by one
            of the following:
                - NAME_DELIMITER_FILE
                - NAME_DELIMITER_CXX
                - NAME_DELIMITER_JAVA
                - NAME_DELIMITER_UNKNOWN
            And then be followed by at least 3 elements separated by
            the delimiter "DELIMITER".
            :return: The NameHierarchy corresponding to the deserialize_name 
        """
        idx = serialized_name.find(NameHierarchy.META_DELIMITER)
        if idx == -1:
            # Invalid meta delimiter

            # @Warning: This is not the same behavior as SourcetrailDB
            # We are returning an exception instead of an empty NameHierarchy
            # return NameHierarchy(NameHierarchy.NAME_DELIMITER_UNKNOWN, None)
            raise DeserializeException()

        elements = list()
        result = NameHierarchy(serialized_name[0: idx], None)

        idx += len(NameHierarchy.META_DELIMITER)
        while idx < len(serialized_name):
            # Read name
            spos = serialized_name.find(NameHierarchy.PART_DELIMITER, idx)
            if spos == -1:
                # Invalid part delimiter

                # @Warning: This is not the same behavior as SourcetrailDB
                # We are returning an exception instead of an empty NameHierarchy
                # return NameHierarchy(NameHierarchy.NAME_DELIMITER_UNKNOWN, None)
                raise DeserializeException()

            name = serialized_name[idx: spos]
            spos += len(NameHierarchy.PART_DELIMITER)

            # Read prefix
            ppos = serialized_name.find(NameHierarchy.SIGNATURE_DELIMITER, spos)
            if ppos == -1:
                # Invalid signature delimiter

                # @Warning: This is not the same behavior as SourcetrailDB
                # We are returning an exception instead of an empty NameHierarchy
                # return NameHierarchy(NameHierarchy.NAME_DELIMITER_UNKNOWN, None)
                raise DeserializeException()

            prefix = serialized_name[spos: ppos]
            ppos += len(NameHierarchy.SIGNATURE_DELIMITER)

            # Read postfix
            npos = serialized_name.find(NameHierarchy.NAME_DELIMITER, ppos)
            if npos == -1:
                postfix = serialized_name[ppos:]
                idx = len(serialized_name)
            else:
                postfix = serialized_name[ppos: npos]
                idx = npos + len(NameHierarchy.NAME_DELIMITER)

            elements.append(NameElement(prefix, name, postfix))

        result._elements = elements
        return result



# -*- coding: utf-8 -*-

#  Copyright 2023 Quarkslab
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Public API of Numbat. Allow to create and manipulate Sourcetrail DB"""

import logging
import os
import sqlite3
from datetime import datetime
from pathlib import Path

from .types import Element, ElementComponent, ElementComponentType, Edge, \
    EdgeType, Node, NodeType, Symbol, SymbolType, File, FileContent, \
    LocalSymbol, SourceLocation, SourceLocationType, Occurrence, Error, \
    NameElement, NameHierarchy

from .db import ComponentAccessDAO, EdgeDAO, ElementComponentDAO, FileDAO, \
    ElementDAO, ErrorDAO, FileContentDAO, LocalSymbolDAO, MetaDAO, \
    NodeDAO, OccurrenceDAO, SourceLocationDAO, SqliteHelper, SymbolDAO

from .exceptions import NoDatabaseOpen, NumbatException


class SourcetrailDB():
    """
    This class implement a wrapper to Sourcetrail internal database,
    it is able to create, edit and delete the underlying sqlite3
    database used by Sourcetrail.
    """

    # Sourcetrail files extension
    SOURCETRAIL_PROJECT_EXT = '.srctrlprj'
    SOURCETRAIL_DB_EXT = '.srctrldb'

    SOURCETRAIL_XML = '\n'.join([
        '<?xml version="1.0" encoding="utf-8" ?>',
        '<config>',
        '   <version>0</version>',
        '</config>'
    ])

    def __init__(self, database: sqlite3.Connection, path: Path, logger: logging.Logger = None) -> None:
        self.database = database
        self.path = path
        if logger is None:
            self.logger = logging.getLogger()
        else:
            self.logger = logger
        self.name_cache = dict()

    # ------------------------------------------------------------------------ #
    # Database file management functions                                       #
    # ------------------------------------------------------------------------ #


    @classmethod
    def __uniformize_path(cls, path: Path | str) -> Path:
        """
        Ensure that the provided path is of type pathlib.Path and has the
        correct suffix, if not add it or cast it.

        :param path: The path to the existing or future database
        :return: a path object
        """
        if type(path) == str:
            path = Path(path)
        if path.suffix != cls.SOURCETRAIL_DB_EXT:
            path = path.with_suffix(cls.SOURCETRAIL_DB_EXT)
        return path.absolute()

    @classmethod
    def exists(cls, path: Path | str) -> bool:
        """
        This method check if there is a Sourcetrail db with the given path.
        If the provided path does not end with the sourcetrail db correct
        suffix. It will be added.

        :param path: The path to test
        :return: a bool
        """
        path = cls.__uniformize_path(path)
        return path.exists()

    @classmethod
    def open(cls, path: Path | str, clear: bool = False) -> 'SourcetrailDB':
        """
        This method allow to open an existing sourcetrail database

        :param path: The path to the existing database
        :param clear: If set to True the database is cleared (Optional)
        :return: the SourcetrailDB object corresponding to the given DB
        """
        path = cls.__uniformize_path(path)
        if not path.exists():
            if path.is_file() or not clear:
                raise FileNotFoundError('%s not found' % str(path))
            return cls.create(path)

        try:
            database = SqliteHelper.connect(str(path))
        except Exception as e:
            raise NumbatException(*e.args)

        obj = SourcetrailDB(database, path)

        if clear:
            # Clear the database if the user ask to
            obj.clear()

        return obj

    @classmethod
    def create(cls, path: Path | str) -> 'SourcetrailDB':
        """
        This method allow to create a sourcetrail database

        :param path: The path to the new database
        :return: the SourcetrailDB object corresponding to the given DB path
        """
        path = cls.__uniformize_path(path)
        if path.exists():
            raise FileExistsError('%s already exists' % str(path))

        try:
            database = SqliteHelper.connect(str(path))
        except Exception as e:
            raise NumbatException(*e.args)

        obj = SourcetrailDB(database, path)
        try:
            obj.__create_sql_tables()
            # add metadata in db
            MetaDAO.new(obj.database, 'storage_version', '25')
            MetaDAO.new(obj.database, 'project_settings', obj.SOURCETRAIL_XML)
            # Create Sourcetrail Project file
            project_file = obj.path.with_suffix(cls.SOURCETRAIL_PROJECT_EXT)
            project_file.write_text(cls.SOURCETRAIL_XML)
            # Commit change to the database so we don't ended up with a half setup DB if an
            # exceptions is raised before the next commit
            obj.commit()
        except Exception as e:
            # They already exists, fail
            obj.close()
            raise NumbatException(*e.args)
        return obj

    def __create_sql_tables(self) -> None:
        """
        This method allow to create all the sql tables needed
        by sourcetrail

        :return: None
        """
        ElementDAO.create_table(self.database)
        ElementComponentDAO.create_table(self.database)
        EdgeDAO.create_table(self.database)
        NodeDAO.create_table(self.database)
        SymbolDAO.create_table(self.database)
        FileDAO.create_table(self.database)
        FileContentDAO.create_table(self.database)
        LocalSymbolDAO.create_table(self.database)
        SourceLocationDAO.create_table(self.database)
        OccurrenceDAO.create_table(self.database)
        ComponentAccessDAO.create_table(self.database)
        ErrorDAO.create_table(self.database)
        MetaDAO.create_table(self.database)

    def commit(self) -> None:
        """
        This method allow to commit changes made to a sourcetrail database.
        Any change made to the database using this API will be lost if not
        committed before closing the database.

        :return: None
        """
        if self.database:
            self.database.commit()
        else:
            raise NoDatabaseOpen()

    def clear(self) -> None:
        """
        Clear all elements present in the database.

        :return: None
        """
        if not self.database:
            raise NoDatabaseOpen()
        ElementDAO.clear(self.database)
        ElementComponentDAO.clear(self.database)
        EdgeDAO.clear(self.database)
        NodeDAO.clear(self.database)
        SymbolDAO.clear(self.database)
        FileDAO.clear(self.database)
        FileContentDAO.clear(self.database)
        LocalSymbolDAO.clear(self.database)
        SourceLocationDAO.clear(self.database)
        OccurrenceDAO.clear(self.database)
        ComponentAccessDAO.clear(self.database)
        ErrorDAO.clear(self.database)

    def close(self) -> None:
        """
        This method allow to close a sourcetrail database.
        The database must be closed after use in order to liberate
        memory and resources allocated for it.

        :return: None
        """
        if self.database:
            self.database.close()
            self.database = None
        else:
            raise NoDatabaseOpen()

    ####################################################################################
    #                        GENERAL SYMBOLS OPERATIONS                                #
    ####################################################################################

    def __add_if_not_existing(self, name: str, type_: NodeType) -> int:
        """
        Create a new node if it does not already exist

        @Warning: This is not the same behavior as SourcetrailDB
        We are not allowing nodes with same serialized_name

        :param name: The serialized_name of the node
        :param type_: The type of the node to insert
        :return: The identifier of the new node or the identifier of
                 the existing one
        """

        if name not in self.name_cache:
            elem = Element()
            elem.id = ElementDAO.new(self.database, elem)

            NodeDAO.new(self.database, Node(
                elem.id,
                type_,
                name
            ))

            self.name_cache[name] = elem.id
            return elem.id
        else:
            return self.name_cache[name]

    def _record_symbol(self, hierarchy: NameHierarchy) -> int:
        """
        Record a new Symbol in the database

        :param hierarchy: The hierarchy of the symbol to insert
        :return: An unique integer that identify the inserted element
        """

        ids = []

        # Add all the nodes needed
        for i in range(0, hierarchy.size()):
            ids.append(self.__add_if_not_existing(
                hierarchy.serialize_range(0, i + 1),
                NodeType.NODE_SYMBOL
            ))

        # Add all the edges between nodes
        if len(ids) > 1:
            parent, children = ids[0], ids[1:]
            for child in children:
                elem = Element()
                elem.id = ElementDAO.new(self.database, elem)

                EdgeDAO.new(self.database, Edge(
                    elem.id,
                    EdgeType.MEMBER,
                    parent,
                    child
                ))

                # Return the id of the last inserted elements
        return ids[-1]

    def _get_symbol(self, hierarchy: NameHierarchy) -> int | None:
        """
        Return the corresponding Symbol from the database

        :param hierarchy: The hierarchy of the symbol to retrieve
        :return: The identifier of the existing symbol or None if the symbol
                 does not exist.
        """

        serialized_name = hierarchy.serialize_name()
        node = NodeDAO.get_by_name(self.database, serialized_name)
        if node:
            return node.id

    def _record_symbol_kind(self, id_: int, type_: NodeType) -> None:
        """
        Set the type of the symbol which is equivalent to setting the
        type of the underlying node.

        :param id_: The identifier of the element
        :param type_: The new type for the symbol
        :return: None
        """

        node = NodeDAO.get(self.database, id_)
        if node:
            node.type = type_
            NodeDAO.update(self.database, node)

    def _record_symbol_definition_kind(self, id_: int, kind: SymbolType) -> None:
        """
        Set the type of definition of the corresponding element

        :param id_: The identifier of the element
        :param kind: The new type for the symbol
        :return: None
        """

        symb = SymbolDAO.get(self.database, id_)
        if symb:
            if symb.definition_kind != kind:
                symb.definition_kind = kind
                SymbolDAO.update(self.database, symb)
        else:
            symb = Symbol(id_, kind)
            SymbolDAO.new(self.database, symb)

    ####################################################################################
    #                                 NODES                                            #
    ####################################################################################

    def __full_record_node(self,
                           name: str,
                           prefix: str,
                           postfix: str,
                           delimiter: str,
                           parent_id: int | None,
                           is_indexed: bool,
                           type_: NodeType) -> int | None:
        """
        Internal function which will be wrapped by all the record_XX methods
        where XX is a node type (class, method, field, etc.). It creates the
        appropriated structures (NameElement, NameHierarchy, etc.) and then insert
        them in the DB. It also handles the typing of the created node plus its
        definition type (explicit or implicit).

        This method also handle the case where the node is the child of an already
        existing node. It automatically creates the hierarchy and so on.

        :param name: The name of the element to insert
        :param prefix: The prefix of the element to insert
        :param postfix: The postfix of the element to insert
        :param delimiter: The delimiter of the element, if the element has a parent,
        it will not be taken into account as the parent delimiter will be used
        :param parent_id: The identifier of the class in which the method is defined.
        :param is_indexed: if the element is explicit or non-indexed
        :param type_: type of the node to add
        :return: The identifier of the new class or None if it could not be inserted
        """
        name_element = NameElement(prefix, name, postfix)
        if parent_id:
            node = NodeDAO.get(self.database, parent_id)
            if not node:
                return
            hierarchy = NameHierarchy.deserialize_name(node.name)
            hierarchy.extend(name_element)
            obj_id = self._record_symbol(hierarchy)
        else:
            obj_id = self._record_symbol(NameHierarchy(delimiter, [name_element]))

        if obj_id:
            self._record_symbol_kind(obj_id, type_)
            if is_indexed:
                self._record_symbol_definition_kind(obj_id, SymbolType.EXPLICIT)
            return obj_id

    def record_symbol_node(self,
                           name: str = '',
                           prefix: str = '',
                           postfix: str = '',
                           delimiter: str = NameHierarchy.NAME_DELIMITER_CXX,
                           parent_id: int = None,
                           is_indexed: bool = True) -> int | None:
        """
        Record a "SYMBOL" symbol into the DB

        :param name: The name of the element to insert
        :param prefix: The prefix of the element to insert
        :param postfix: The postfix of the element to insert
        :param delimiter: The delimiter of the element, if the element has a parent,
        it will not be taken into account as the parent delimiter will be used
        :param parent_id: The identifier of the class in which the method is defined.
        :param is_indexed: if the element is explicit or non-indexed
        :return: The identifier of the new class or None if it could not be inserted
        """

        return self.__full_record_node(name, prefix, postfix, delimiter,
                                       parent_id, is_indexed, NodeType.NODE_SYMBOL)

    def record_type_node(self,
                         name: str = '',
                         prefix: str = '',
                         postfix: str = '',
                         delimiter: str = NameHierarchy.NAME_DELIMITER_CXX,
                         parent_id: int = None,
                         is_indexed: bool = True) -> int | None:
        """
        Record a TYPE symbol into the DB

        :param name: The name of the element to insert
        :param prefix: The prefix of the element to insert
        :param postfix: The postfix of the element to insert
        :param delimiter: The delimiter of the element, if the element has a parent,
        it will not be taken into account as the parent delimiter will be used
        :param parent_id: The identifier of the class in which the method is defined.
        :param is_indexed: if the element is explicit or non-indexed
        :return: The identifier of the new class or None if it could not be inserted
        """

        return self.__full_record_node(name, prefix, postfix, delimiter,
                                       parent_id, is_indexed, NodeType.NODE_TYPE)

    def record_buitin_type_node(self,
                                name: str = '',
                                prefix: str = '',
                                postfix: str = '',
                                delimiter: str = NameHierarchy.NAME_DELIMITER_CXX,
                                parent_id: int = None,
                                is_indexed: bool = True) -> int | None:
        """
        Record a BUILTIN_TYPE symbol into the DB

        :param name: The name of the element to insert
        :param prefix: The prefix of the element to insert
        :param postfix: The postfix of the element to insert
        :param delimiter: The delimiter of the element, if the element has a parent,
        it will not be taken into account as the parent delimiter will be used
        :param parent_id: The identifier of the class in which the method is defined.
        :param is_indexed: if the element is explicit or non-indexed
        :return: The identifier of the new class or None if it could not be inserted
        """

        return self.__full_record_node(name, prefix, postfix, delimiter,
                                       parent_id, is_indexed, NodeType.NODE_BUILTIN_TYPE)

    def record_module(self,
                      name: str = '',
                      prefix: str = '',
                      postfix: str = '',
                      delimiter: str = NameHierarchy.NAME_DELIMITER_CXX,
                      parent_id: int = None,
                      is_indexed: bool = True) -> int | None:
        """
        Record a MODULE symbol into the DB

        :param name: The name of the element to insert
        :param prefix: The prefix of the element to insert
        :param postfix: The postfix of the element to insert
        :param delimiter: The delimiter of the element, if the element has a parent,
        it will not be taken into account as the parent delimiter will be used
        :param parent_id: The identifier of the class in which the method is defined.
        :param is_indexed: if the element is explicit or non-indexed
        :return: The identifier of the new class or None if it could not be inserted
        """

        return self.__full_record_node(name, prefix, postfix, delimiter,
                                       parent_id, is_indexed, NodeType.NODE_MODULE)

    def record_namespace(self,
                         name: str = '',
                         prefix: str = '',
                         postfix: str = '',
                         delimiter: str = NameHierarchy.NAME_DELIMITER_CXX,
                         parent_id: int = None,
                         is_indexed: bool = True) -> int | None:
        """
        Record a NAMESPACE symbol into the DB

        :param name: The name of the element to insert
        :param prefix: The prefix of the element to insert
        :param postfix: The postfix of the element to insert
        :param delimiter: The delimiter of the element, if the element has a parent,
        it will not be taken into account as the parent delimiter will be used
        :param parent_id: The identifier of the class in which the method is defined.
        :param is_indexed: if the element is explicit or non-indexed
        :return: The identifier of the new class or None if it could not be inserted
        """

        return self.__full_record_node(name, prefix, postfix, delimiter,
                                       parent_id, is_indexed, NodeType.NODE_NAMESPACE)

    def record_package(self,
                       name: str = '',
                       prefix: str = '',
                       postfix: str = '',
                       delimiter: str = NameHierarchy.NAME_DELIMITER_CXX,
                       parent_id: int = None,
                       is_indexed: bool = True) -> int | None:
        """
        Record a PACKAGE symbol into the DB

        :param name: The name of the element to insert
        :param prefix: The prefix of the element to insert
        :param postfix: The postfix of the element to insert
        :param delimiter: The delimiter of the element, if the element has a parent,
        it will not be taken into account as the parent delimiter will be used
        :param parent_id: The identifier of the class in which the method is defined.
        :param is_indexed: if the element is explicit or non-indexed
        :return: The identifier of the new class or None if it could not be inserted
        """

        return self.__full_record_node(name, prefix, postfix, delimiter,
                                       parent_id, is_indexed, NodeType.NODE_PACKAGE)

    def record_struct(self,
                      name: str = '',
                      prefix: str = '',
                      postfix: str = '',
                      delimiter: str = NameHierarchy.NAME_DELIMITER_CXX,
                      parent_id: int = None,
                      is_indexed: bool = True) -> int | None:
        """
        Record a STRUCT symbol into the DB

        :param name: The name of the element to insert
        :param prefix: The prefix of the element to insert
        :param postfix: The postfix of the element to insert
        :param delimiter: The delimiter of the element, if the element has a parent,
        it will not be taken into account as the parent delimiter will be used
        :param parent_id: The identifier of the class in which the method is defined.
        :param is_indexed: if the element is explicit or non-indexed
        :return: The identifier of the new class or None if it could not be inserted
        """

        return self.__full_record_node(name, prefix, postfix, delimiter,
                                       parent_id, is_indexed, NodeType.NODE_STRUCT)

    def record_class(self,
                     name: str = '',
                     prefix: str = '',
                     postfix: str = '',
                     delimiter: str = NameHierarchy.NAME_DELIMITER_CXX,
                     parent_id: int = None,
                     is_indexed: bool = True) -> int | None:
        """
        Record a CLASS symbol into the DB

        :param name: The name of the element to insert
        :param prefix: The prefix of the element to insert
        :param postfix: The postfix of the element to insert
        :param delimiter: The delimiter of the element, if the element has a parent,
        it will not be taken into account as the parent delimiter will be used
        :param parent_id: The identifier of the class in which the method is defined.
        :param is_indexed: if the element is explicit or non-indexed
        :return: The identifier of the new class or None if it could not be inserted
        """

        return self.__full_record_node(name, prefix, postfix, delimiter,
                                       parent_id, is_indexed, NodeType.NODE_CLASS)

    def record_interface(self,
                         name: str = '',
                         prefix: str = '',
                         postfix: str = '',
                         delimiter: str = NameHierarchy.NAME_DELIMITER_CXX,
                         parent_id: int = None,
                         is_indexed: bool = True) -> int | None:
        """
        Record a INTERFACE symbol into the DB

        :param name: The name of the element to insert
        :param prefix: The prefix of the element to insert
        :param postfix: The postfix of the element to insert
        :param delimiter: The delimiter of the element, if the element has a parent,
        it will not be taken into account as the parent delimiter will be used
        :param parent_id: The identifier of the class in which the method is defined.
        :param is_indexed: if the element is explicit or non-indexed
        :return: The identifier of the new class or None if it could not be inserted
        """

        return self.__full_record_node(name, prefix, postfix, delimiter,
                                       parent_id, is_indexed, NodeType.NODE_INTERFACE)

    def record_annotation(self,
                          name: str = '',
                          prefix: str = '',
                          postfix: str = '',
                          delimiter: str = NameHierarchy.NAME_DELIMITER_CXX,
                          parent_id: int = None,
                          is_indexed: bool = True) -> int | None:
        """
        Record a ANNOTATION symbol into the DB

        :param name: The name of the element to insert
        :param prefix: The prefix of the element to insert
        :param postfix: The postfix of the element to insert
        :param delimiter: The delimiter of the element, if the element has a parent,
        it will not be taken into account as the parent delimiter will be used
        :param parent_id: The identifier of the class in which the method is defined.
        :param is_indexed: if the element is explicit or non-indexed
        :return: The identifier of the new class or None if it could not be inserted
        """

        return self.__full_record_node(name, prefix, postfix, delimiter,
                                       parent_id, is_indexed, NodeType.NODE_ANNOTATION)

    def record_global_variable(self,
                               name: str = '',
                               prefix: str = '',
                               postfix: str = '',
                               delimiter: str = NameHierarchy.NAME_DELIMITER_CXX,
                               parent_id: int = None,
                               is_indexed: bool = True) -> int | None:
        """
        Record a GLOBAL_VARIABLE symbol into the DB

        :param name: The name of the element to insert
        :param prefix: The prefix of the element to insert
        :param postfix: The postfix of the element to insert
        :param delimiter: The delimiter of the element, if the element has a parent,
        it will not be taken into account as the parent delimiter will be used
        :param parent_id: The identifier of the class in which the method is defined.
        :param is_indexed: if the element is explicit or non-indexed
        :return: The identifier of the new class or None if it could not be inserted
        """

        return self.__full_record_node(name, prefix, postfix, delimiter,
                                       parent_id, is_indexed, NodeType.NODE_GLOBAL_VARIABLE)

    def record_field(self,
                     name: str = '',
                     prefix: str = '',
                     postfix: str = '',
                     delimiter: str = NameHierarchy.NAME_DELIMITER_CXX,
                     parent_id: int = None,
                     is_indexed: bool = True) -> int | None:
        """
        Record a FIELD symbol into the DB

        :param name: The name of the element to insert
        :param prefix: The prefix of the element to insert
        :param postfix: The postfix of the element to insert
        :param delimiter: The delimiter of the element, if the element has a parent,
        it will not be taken into account as the parent delimiter will be used
        :param parent_id: The identifier of the class in which the method is defined.
        :param is_indexed: if the element is explicit or non-indexed
        :return: The identifier of the new class or None if it could not be inserted
        """

        return self.__full_record_node(name, prefix, postfix, delimiter,
                                       parent_id, is_indexed, NodeType.NODE_FIELD)

    def record_function(self,
                        name: str = '',
                        prefix: str = '',
                        postfix: str = '',
                        delimiter: str = NameHierarchy.NAME_DELIMITER_CXX,
                        parent_id: int = None,
                        is_indexed: bool = True) -> int | None:
        """
        Record a FUNCTION symbol into the DB

        :param name: The name of the element to insert
        :param prefix: The prefix of the element to insert
        :param postfix: The postfix of the element to insert
        :param delimiter: The delimiter of the element, if the element has a parent,
        it will not be taken into account as the parent delimiter will be used
        :param parent_id: The identifier of the class in which the method is defined.
        :param is_indexed: if the element is explicit or non-indexed
        :return: The identifier of the new class or None if it could not be inserted
        """

        return self.__full_record_node(name, prefix, postfix, delimiter,
                                       parent_id, is_indexed, NodeType.NODE_FUNCTION)

    def record_method(self,
                      name: str = '',
                      prefix: str = '',
                      postfix: str = '',
                      delimiter: str = NameHierarchy.NAME_DELIMITER_CXX,
                      parent_id: int = None,
                      is_indexed: bool = True) -> int | None:
        """
        Record a METHOD symbol into the DB

        :param name: The name of the element to insert
        :param prefix: The prefix of the element to insert
        :param postfix: The postfix of the element to insert
        :param delimiter: The delimiter of the element, if the element has a parent,
        it will not be taken into account as the parent delimiter will be used
        :param parent_id: The identifier of the class in which the method is defined.
        :param is_indexed: if the element is explicit or non-indexed
        :return: The identifier of the new class or None if it could not be inserted
        """

        return self.__full_record_node(name, prefix, postfix, delimiter,
                                       parent_id, is_indexed, NodeType.NODE_METHOD)

    def record_enum(self,
                    name: str = '',
                    prefix: str = '',
                    postfix: str = '',
                    delimiter: str = NameHierarchy.NAME_DELIMITER_CXX,
                    parent_id: int = None,
                    is_indexed: bool = True) -> int | None:
        """
        Record a ENUM symbol into the DB

        :param name: The name of the element to insert
        :param prefix: The prefix of the element to insert
        :param postfix: The postfix of the element to insert
        :param delimiter: The delimiter of the element, if the element has a parent,
        it will not be taken into account as the parent delimiter will be used
        :param parent_id: The identifier of the class in which the method is defined.
        :param is_indexed: if the element is explicit or non-indexed
        :return: The identifier of the new class or None if it could not be inserted
        """

        return self.__full_record_node(name, prefix, postfix, delimiter,
                                       parent_id, is_indexed, NodeType.NODE_ENUM)

    def record_enum_constant(self,
                             name: str = '',
                             prefix: str = '',
                             postfix: str = '',
                             delimiter: str = NameHierarchy.NAME_DELIMITER_CXX,
                             parent_id: int = None,
                             is_indexed: bool = True) -> int | None:
        """
        Record a ENUM_CONSTANT symbol into the DB

        :param name: The name of the element to insert
        :param prefix: The prefix of the element to insert
        :param postfix: The postfix of the element to insert
        :param delimiter: The delimiter of the element, if the element has a parent,
        it will not be taken into account as the parent delimiter will be used
        :param parent_id: The identifier of the class in which the method is defined.
        :param is_indexed: if the element is explicit or non-indexed
        :return: The identifier of the new class or None if it could not be inserted
        """

        return self.__full_record_node(name, prefix, postfix, delimiter,
                                       parent_id, is_indexed, NodeType.NODE_ENUM_CONSTANT)

    def record_typedef_node(self,
                            name: str = '',
                            prefix: str = '',
                            postfix: str = '',
                            delimiter: str = NameHierarchy.NAME_DELIMITER_CXX,
                            parent_id: int = None,
                            is_indexed: bool = True) -> int | None:
        """
        Record a TYPEDEF symbol into the DB

        :param name: The name of the element to insert
        :param prefix: The prefix of the element to insert
        :param postfix: The postfix of the element to insert
        :param delimiter: The delimiter of the element, if the element has a parent,
        it will not be taken into account as the parent delimiter will be used
        :param parent_id: The identifier of the class in which the method is defined.
        :param is_indexed: if the element is explicit or non-indexed
        :return: The identifier of the new class or None if it could not be inserted
        """

        return self.__full_record_node(name, prefix, postfix, delimiter,
                                       parent_id, is_indexed, NodeType.NODE_TYPEDEF)

    def record_type_parameter_node(self,
                                   name: str = '',
                                   prefix: str = '',
                                   postfix: str = '',
                                   delimiter: str = NameHierarchy.NAME_DELIMITER_CXX,
                                   parent_id: int = None,
                                   is_indexed: bool = True) -> int | None:
        """
        Record a TYPE_PARAMETER symbol into the DB

        :param name: The name of the element to insert
        :param prefix: The prefix of the element to insert
        :param postfix: The postfix of the element to insert
        :param delimiter: The delimiter of the element, if the element has a parent,
        it will not be taken into account as the parent delimiter will be used
        :param parent_id: The identifier of the class in which the method is defined.
        :param is_indexed: if the element is explicit or non-indexed
        :return: The identifier of the new class or None if it could not be inserted
        """

        return self.__full_record_node(name, prefix, postfix, delimiter,
                                       parent_id, is_indexed, NodeType.NODE_TYPE_PARAMETER)

    def record_macro(self,
                     name: str = '',
                     prefix: str = '',
                     postfix: str = '',
                     delimiter: str = NameHierarchy.NAME_DELIMITER_CXX,
                     parent_id: int = None,
                     is_indexed: bool = True) -> int | None:
        """
        Record a MACRO symbol into the DB

        :param name: The name of the element to insert
        :param prefix: The prefix of the element to insert
        :param postfix: The postfix of the element to insert
        :param delimiter: The delimiter of the element, if the element has a parent,
        it will not be taken into account as the parent delimiter will be used
        :param parent_id: The identifier of the class in which the method is defined.
        :param is_indexed: if the element is explicit or non-indexed
        :return: The identifier of the new class or None if it could not be inserted
        """

        return self.__full_record_node(name, prefix, postfix, delimiter,
                                       parent_id, is_indexed, NodeType.NODE_MACRO)

    def record_union(self,
                     name: str = '',
                     prefix: str = '',
                     postfix: str = '',
                     delimiter: str = NameHierarchy.NAME_DELIMITER_CXX,
                     parent_id: int = None,
                     is_indexed: bool = True) -> int | None:
        """
        Record a UNION symbol into the DB

        :param name: The name of the element to insert
        :param prefix: The prefix of the element to insert
        :param postfix: The postfix of the element to insert
        :param delimiter: The delimiter of the element, if the element has a parent,
        it will not be taken into account as the parent delimiter will be used
        :param parent_id: The identifier of the class in which the method is defined.
        :param is_indexed: if the element is explicit or non-indexed
        :return: The identifier of the new class or None if it could not be inserted
        """

        return self.__full_record_node(name, prefix, postfix, delimiter,
                                       parent_id, is_indexed, NodeType.NODE_UNION)

    ####################################################################################
    #                               REFERENCES                                         #
    ####################################################################################

    # Add new references

    def _record_reference(self, source_id: int, dest_id: int, type_: EdgeType) -> int:
        """
        Add a new reference (an edge) between two elements

        :param source_id: The source identifier of the reference
        :param dest_id: The destination identifier of the reference
        :param type_: The type of reference to add
        :return: None
        """

        elem = Element()
        elem.id = ElementDAO.new(self.database, elem)

        EdgeDAO.new(self.database, Edge(elem.id, type_, source_id, dest_id))

        return elem.id

    def record_ref_member(self, source_id: int, dest_id: int) -> int:
        """
        Add a member reference (aka an edge) between two elements

        :param source_id: The source identifier
        :param dest_id: The destination identifier
        :return: the reference id
        """
        return self._record_reference(source_id, dest_id, type_=EdgeType.MEMBER)

    def record_ref_type_usage(self, source_id: int, dest_id: int) -> int:
        """
        Add a TYPE_USAGE reference (aka an edge) between two elements

        :param source_id: The source identifier
        :param dest_id: The destination identifier
        :return: the reference id
        """
        return self._record_reference(source_id, dest_id, type_=EdgeType.TYPE_USAGE)

    def record_ref_usage(self, source_id: int, dest_id: int) -> int:
        """
        Add a USAGE reference (aka an edge) between two elements

        :param source_id: The source identifier
        :param dest_id: The destination identifier
        :return: the reference id
        """
        return self._record_reference(source_id, dest_id, type_=EdgeType.USAGE)

    def record_ref_call(self, source_id: int, dest_id: int) -> int:
        """
        Add a CALL reference (aka an edge) between two elements

        :param source_id: The source identifier
        :param dest_id: The destination identifier
        :return: the reference id
        """
        return self._record_reference(source_id, dest_id, type_=EdgeType.CALL)

    def record_ref_inheritance(self, source_id: int, dest_id: int) -> int:
        """
        Add an INHERITANCE reference (aka an edge) between two elements

        :param source_id: The source identifier
        :param dest_id: The destination identifier
        :return: the reference id
        """
        return self._record_reference(source_id, dest_id, type_=EdgeType.INHERITANCE)

    def record_ref_override(self, source_id: int, dest_id: int) -> int:
        """
        Add an OVERRIDE reference (aka an edge) between two elements

        :param source_id: The source identifier
        :param dest_id: The destination identifier
        :return: the reference id
        """
        return self._record_reference(source_id, dest_id, type_=EdgeType.OVERRIDE)

    def record_ref_type_argument(self, source_id: int, dest_id: int) -> int:
        """
        Add a TYPE_ARGUMENT reference (aka an edge) between two elements

        :param source_id: The source identifier
        :param dest_id: The destination identifier
        :return: the reference id
        """
        return self._record_reference(source_id, dest_id, type_=EdgeType.TYPE_ARGUMENT)

    def record_ref_template_specialization(self, source_id: int, dest_id: int) -> int:
        """
        Add a TEMPLATE_SPECIALIZATION reference (aka an edge) between two elements

        :param source_id: The source identifier
        :param dest_id: The destination identifier
        :return: the reference id
        """
        return self._record_reference(source_id, dest_id, type_=EdgeType.TEMPLATE_SPECIALIZATION)

    def record_ref_include(self, source_id: int, dest_id: int) -> int:
        """
        Add an INCLUDE reference (aka an edge) between two elements

        :param source_id: The source identifier
        :param dest_id: The destination identifier
        :return: the reference id
        """
        return self._record_reference(source_id, dest_id, type_=EdgeType.INCLUDE)

    def record_ref_import(self, source_id: int, dest_id: int) -> int:
        """
        Add an import reference (aka an edge) between two elements

        :param source_id: The source identifier (who imports)
        :param dest_id: The destination identifier (who is imported)
        :return: the reference id
        """
        return self._record_reference(source_id, dest_id, type_=EdgeType.IMPORT)

    def record_ref_bundled_edges(self, source_id: int, dest_id: int) -> int:
        """
        Add a BUNDLED_EDGES reference (aka an edge) between two elements

        :param source_id: The source identifier
        :param dest_id: The destination identifier
        :return: the reference id
        """
        return self._record_reference(source_id, dest_id, type_=EdgeType.BUNDLED_EDGES)

    def record_ref_macro_usage(self, source_id: int, dest_id: int) -> int:
        """
        Add a MACRO_USAGE reference (aka an edge) between two elements

        :param source_id: The source identifier
        :param dest_id: The destination identifier
        :return: the reference id
        """
        return self._record_reference(source_id, dest_id, type_=EdgeType.MACRO_USAGE)

    def record_ref_annotation_usage(self, source_id: int, dest_id: int) -> int:
        """
        Add an ANNOTATION_USAGE reference (aka an edge) between two elements

        :param source_id: The source identifier
        :param dest_id: The destination identifier
        :return: the reference id
        """
        return self._record_reference(source_id, dest_id, type_=EdgeType.ANNOTATION_USAGE)

    def record_reference_to_unsolved_symbol(self, symbol_id: int, reference_type: EdgeType,
                                            file_id: int, start_line: int, start_column: int, end_line: int,
                                            end_column: int) -> int:
        """
        Record a reference to an unsolved symbol.

        :param symbol_id: The identifier of the symbol
        :param reference_type: The type of reference
        :param file_id: The identifier of the source file in which the symbol is located
        :param start_line: The line at which the element starts.
        :param start_column: The column at which the element starts.
        :param end_line: The line at which the element ends.
        :param end_column: The line at which the element ends.
        :return: The identifier of the new reference
        """

        # Don't blame me, it's done like this in sourcetrail source code
        hierarchy = NameHierarchy(
            NameHierarchy.NAME_DELIMITER_UNKNOWN,
            [NameElement(
                '',
                'unsolved symbol',
                ''
            )]
        )

        # Insert the new node
        unsolved_symbol_id = self._record_symbol(hierarchy)

        # Add a new edge
        elem = Element()
        elem.id = ElementDAO.new(self.database, elem)

        reference_id = EdgeDAO.new(self.database, Edge(
            elem.id,
            reference_type,
            symbol_id,
            unsolved_symbol_id
        ))

        # Add the new source location
        self.__record_source_location(
            reference_id,
            file_id,
            start_line,
            start_column,
            end_line,
            end_column,
            SourceLocationType.UNSOLVED
        )

        # Return edge id
        return reference_id

    # Modify existing references
    def record_reference_is_ambiguous(self, reference_id: int) -> None:
        """
        Add an indication in the database to tell that the reference is ambiguous

        :param reference_id: the identifier of the reference
        :return: None
        """

        ElementComponentDAO.new(self.database, ElementComponent(
            0,
            reference_id,
            ElementComponentType.IS_AMBIGUOUS,
            ''
        ))

    ####################################################################################
    #                           SOURCE CODE MANIPULATION                               #
    ####################################################################################

    def record_file(self, path: Path, indexed: bool = True) -> int:
        """
        Record a source file in the database

        :param path: The path to the existing source file
        :param indexed: A boolean that indicates if the source file
                        was indexed by the parser
        :return: The identifier of the inserted file
        """

        if not path.exists() or not path.is_file():
            raise FileNotFoundError()

        # Create a new name hierarchy
        hierarchy = NameHierarchy(
            NameHierarchy.NAME_DELIMITER_FILE,
            [NameElement(
                '',
                str(path.absolute()),
                ''
            )]
        )

        # Retrieve the modification date in the correct format
        modification_time = datetime.fromtimestamp(
            os.path.getmtime(path)
        ).strftime("%Y-%m-%d %H:%M:%S")

        # Read the file
        lines = []
        if indexed:
            lines = open(path, 'r').readlines()

        # Insert a new node
        elem_id = self.__add_if_not_existing(
            hierarchy.serialize_name(),
            NodeType.NODE_FILE
        )

        # Insert a new file
        FileDAO.new(
            self.database,
            File(
                elem_id,
                str(path.absolute()),
                '',  # Empty language identifier for now
                modification_time,
                indexed,
                True,
                len(lines)
            )
        )

        if indexed:
            # Insert a new filecontent
            FileContentDAO.new(
                self.database,
                FileContent(elem_id, ''.join(lines))
            )

        # Return the newly created element id
        return elem_id

    def record_file_language(self, id_: int, language: str) -> None:
        """
            Set the language of an existing file inside the database
            :param id_: The identifier of the file
            :param language: A string that indicate the programming language of the file
            :return: None
        """

        file = FileDAO.get(self.database, id_)
        if file:
            file.language = language
            FileDAO.update(self.database, file)

    def __record_source_location(self, symbol_id: int, file_id: int,
                                 start_line: int, start_column: int, end_line: int, end_column: int,
                                 type_: SourceLocationType) -> None:
        """
        Wrapper for all the record_*_location

        :param symbol_id: The identifier of the symbol
        :param file_id: The identifier of the source file in which the symbol is located
        :param start_line: The line at which the element starts.
        :param start_column: The column at which the element starts.
        :param end_line: The line at which the element ends.
        :param end_column: The line at which the element ends.
        :param type_: The type of the source location.
        :return: None
        """

        # First add a new source location entry
        loc_id = SourceLocationDAO.new(self.database, SourceLocation(
            0,
            file_id,
            start_line,
            start_column,
            end_line,
            end_column,
            type_
        ))

        # Now add an occurrence that refer to this location
        OccurrenceDAO.new(self.database, Occurrence(
            symbol_id, loc_id
        ))

    def record_symbol_location(self, symbol_id: int, file_id: int, start_line: int,
                               start_column: int, end_line: int, end_column: int) -> None:

        """
        Record a new source location of type TOKEN

        :param symbol_id: The identifier of the symbol
        :param file_id: The identifier of the source file in which the symbol is located
        :param start_line: The line at which the element starts.
        :param start_column: The column at which the element starts.
        :param end_line: The line at which the element ends.
        :param end_column: The line at which the element ends.
        :return: None
        """

        self.__record_source_location(
            symbol_id,
            file_id,
            start_line,
            start_column,
            end_line,
            end_column,
            SourceLocationType.TOKEN
        )

    def record_symbol_scope_location(self, symbol_id: int, file_id: int, start_line: int,
                                     start_column: int, end_line: int, end_column: int) -> None:
        """
        Record a new source location of type SCOPE

        :param symbol_id: The identifier of the symbol
        :param file_id: The identifier of the source file in which the symbol is located
        :param start_line: The line at which the element starts.
        :param start_column: The column at which the element starts.
        :param end_line: The line at which the element ends.
        :param end_column: The line at which the element ends.
        :return: None
        """

        self.__record_source_location(
            symbol_id,
            file_id,
            start_line,
            start_column,
            end_line,
            end_column,
            SourceLocationType.SCOPE
        )

    def record_symbol_signature_location(self, symbol_id: int, file_id: int, start_line: int,
                                         start_column: int, end_line: int, end_column: int) -> None:
        """
        Record a new source location of type SCOPE

        :param symbol_id: The identifier of the symbol
        :param file_id: The identifier of the source file in which the symbol is located
        :param start_line: The line at which the element starts.
        :param start_column: The column at which the element starts.
        :param end_line: The line at which the element ends.
        :param end_column: The line at which the element ends.
        :return: None
        """

        self.__record_source_location(
            symbol_id,
            file_id,
            start_line,
            start_column,
            end_line,
            end_column,
            SourceLocationType.SIGNATURE
        )

    def record_reference_location(self, reference_id: int, file_id: int, start_line: int,
                                  start_column: int, end_line: int, end_column: int) -> None:
        """
        Record a new reference location of type TOKEN

        :param reference_id: The reference identifier
        :param file_id: The identifier of the source file in which the symbol is located
        :param start_line: The line at which the element starts.
        :param start_column: The column at which the element starts.
        :param end_line: The line at which the element ends.
        :param end_column: The line at which the element ends.
        :return: None
        """

        self.__record_source_location(
            reference_id,
            file_id,
            start_line,
            start_column,
            end_line,
            end_column,
            SourceLocationType.TOKEN
        )

    def record_qualifier_location(self, symbol_id: int, file_id: int, start_line: int,
                                  start_column: int, end_line: int, end_column: int) -> None:
        """
        Record a new source location of type QUALIFIER

        :param symbol_id: The identifier of the symbol
        :param file_id: The identifier of the source file in which the symbol is located
        :param start_line: The line at which the element starts.
        :param start_column: The column at which the element starts.
        :param end_line: The line at which the element ends.
        :param end_column: The line at which the element ends.
        :return: None
        """

        self.__record_source_location(
            symbol_id,
            file_id,
            start_line,
            start_column,
            end_line,
            end_column,
            SourceLocationType.QUALIFIER
        )

    def record_local_symbol(self, name: str) -> int:
        """
        Record a new local symbol

        :param name: The name of the new local symbol
        :return: The identifier of the new local symbol
        """

        # Check that the symbol does not already exist
        local = LocalSymbolDAO.get_from_name(self.database, name)
        if not local:
            # Insert a new local symbol
            elem = Element()
            elem.id = ElementDAO.new(self.database, elem)
            local = LocalSymbol(elem.id, name)
            LocalSymbolDAO.new(self.database, local)

        return local.id

    def record_local_symbol_location(self, symbol_id: int, file_id: int, start_line: int,
                                     start_column: int, end_line: int, end_column: int) -> None:
        """
        Record a new source location of type LOCAL_SYMBOL

        :param symbol_id: The identifier of the symbol
        :param file_id: The identifier of the source file in which the symbol is located
        :param start_line: The line at which the element starts.
        :param start_column: The column at which the element starts.
        :param end_line: The line at which the element ends.
        :param end_column: The line at which the element ends.
        :return: None
        """

        self.__record_source_location(
            symbol_id,
            file_id,
            start_line,
            start_column,
            end_line,
            end_column,
            SourceLocationType.LOCAL_SYMBOL
        )

    def record_atomic_source_range(self, symbol_id: int, file_id: int, start_line: int,
                                   start_column: int, end_line: int, end_column: int) -> None:
        """
        Record a new source location of type ATOMIC_RANGE

        :param symbol_id: The identifier of the symbol
        :param file_id: The identifier of the source file in which the symbol is located
        :param start_line: The line at which the element starts.
        :param start_column: The column at which the element starts.
        :param end_line: The line at which the element ends.
        :param end_column: The line at which the element ends.
        :return: None
    """

        self.__record_source_location(
            symbol_id,
            file_id,
            start_line,
            start_column,
            end_line,
            end_column,
            SourceLocationType.ATOMIC_RANGE
        )

    def record_error(self, msg: str, fatal: bool, file_id: int, start_line: int,
                     start_column: int, end_line: int, end_column: int) -> None:
        """
        Record a new indexer error in the database

        :param msg: The description of the error
        :param fatal: A boolean indicating if the error stop the execution of the parser
        :param file_id: The identifier of the source file being parsed
        :param start_line: The line at which the element starts.
        :param start_column: The column at which the element starts.
        :param end_line: The line at which the element ends.
        :param end_column: The line at which the element ends.
        :return: None
        """

        # Add a new error
        elem = Element()
        elem.id = ElementDAO.new(self.database, elem)

        error_id = ErrorDAO.new(self.database, Error(
            elem.id,
            msg,
            fatal,
            True,
            ''
        ))
        self.__record_source_location(
            error_id,
            file_id,
            start_line,
            start_column,
            end_line,
            end_column,
            SourceLocationType.INDEXER_ERROR
        )
